#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import time
import stat
import errno
import shutil

from pathlib import Path

from .database import DataBase
from .utils import (InterruptError, AbortedError, SkippedError)
from .debug_print import (debug_print, debug_pprint)

from .fallocate import *


def rnr_copyfile(cur_file, cur_target, file_size, block_size, resume, info, timers, fd, q, ev_skip, ev_suspend, ev_interrupt, ev_abort):
	with open(cur_file, 'rb') as fh:
		if resume:
			target_fd = os.open(cur_target, os.O_WRONLY | os.O_DSYNC, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)
		else:
			target_fd = os.open(cur_target, os.O_CREAT | os.O_EXCL | os.O_TRUNC | os.O_WRONLY | os.O_DSYNC, stat.S_IRUSR | stat.S_IWUSR | stat.S_IRGRP | stat.S_IWGRP | stat.S_IROTH | stat.S_IWOTH)

		try:
			if resume:
				bytes_written = os.fstat(target_fd).st_size
				pos = max((int(bytes_written / block_size) - 1) * block_size, 0)
				os.lseek(target_fd, pos, os.SEEK_SET)
				fh.seek(pos)
				info['cur_bytes'] += pos
				info['bytes'] += pos
			else:
				try:
					fallocate(target_fd, FALLOC_FL_KEEP_SIZE, 0, file_size)
				except OSError:
					pass

			while True:
				if ev_interrupt.is_set():
					raise InterruptError()

				t1 = time.monotonic()
				ev_suspend.wait()
				t2 = time.monotonic()
				dt = round(t2 - t1)
				timers['cur_start'] += dt
				timers['start'] += dt

				if ev_abort.is_set():
					raise AbortedError()

				if ev_skip.is_set():
					ev_skip.clear()
					raise SkippedError('ev_skip')

				buf = fh.read(block_size)
				if not buf:
					break

				buffer_length = len(buf)
				bytes_written = 0
				while bytes_written < buffer_length:
					bytes_written += os.write(target_fd, buf[bytes_written:])

				info['cur_bytes'] += bytes_written
				info['bytes'] += bytes_written
				now = time.monotonic()
				if (now - timers['last_write']) > 0.05:
					timers['last_write'] = now
					info['cur_time'] = int(round(now - timers['cur_start']))
					info['time'] = int(round(now - timers['start']))
					q.put(info.copy())
					try:
						os.write(fd, b'\n')
					except OSError:
						pass
		finally:
			os.close(target_fd)

def rnr_cpmv(mode, files, cwd, dest, on_conflict, fd, q, ev_skip, ev_suspend, ev_interrupt, ev_abort, ev_nodb, dbfile, job_id):
	if dbfile:
		db = DataBase(dbfile)

	file_list = sorted(files, key=lambda x: x['file'].replace(os.sep, '\0'))
	error_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'ERROR']
	skipped_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'SKIPPED']
	completed_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'DONE']
	aborted_list = []

	dest = Path(dest)

	try:
		block_size = max(dest.lstat().st_blksize, 131072)
	except (OSError, AttributeError):
		block_size = 131072

	info = {
		'cur_source': '',
		'cur_target': '',
		'cur_size': 0,
		'cur_bytes': 0,
		'cur_time': 0,
		'files': 0,
		'bytes': 0,
		'time': 0,
	}

	timers = {}

	if dbfile:
		dir_list = db.get_dir_list(job_id)
		rename_dir_stack = db.get_rename_dir_stack(job_id)
		skip_dir_stack = db.get_skip_dir_stack(job_id)
		replace_first_path = db.get_replace_first_path(job_id)
	else:
		dir_list = []
		rename_dir_stack = []
		skip_dir_stack = []
		replace_first_path = None

	if replace_first_path is None:
		if dest.is_dir():
			replace_first_path = False
		else:
			replace_first_path = True

		if dbfile:
			db.set_replace_first_path(job_id, replace_first_path)

	total_bytes = 0
	timers['start'] = time.monotonic()
	timers['last_write'] = timers['start']
	for i_file, file in enumerate(file_list):
		try:
			if ev_interrupt.is_set():
				raise InterruptError()

			if dbfile and ev_nodb.is_set():
				db.delete_job(job_id)
				del db
				dbfile = None

			if file['status'] in ('DONE', 'ERROR', 'SKIPPED'):
				raise SkippedError('no_log')

			timers['cur_start'] = time.monotonic()

			t1 = time.monotonic()
			ev_suspend.wait()
			t2 = time.monotonic()
			dt = round(t2 - t1)
			timers['cur_start'] += dt
			timers['start'] += dt

			cur_file = Path(file['file'])
			rel_file = cur_file.relative_to(cwd)

			if replace_first_path:
				cur_target = dest / os.sep.join(rel_file.parts[1:])
			else:
				cur_target = dest / rel_file

			skip_dir_stack_changed = False
			skip_dir = False
			while skip_dir_stack:
				dir_to_skip = skip_dir_stack[-1]
				if dir_to_skip in cur_file.parents:
					skip_dir = True
					break
				else:
					skip_dir_stack.pop()
					skip_dir_stack_changed = True

			if skip_dir_stack_changed and dbfile:
				db.set_skip_dir_stack(job_id, skip_dir_stack)

			rename_dir_stack_changed = False
			(old_target, new_target) = (None, None)
			while rename_dir_stack:
				(old_target, new_target) = rename_dir_stack[-1]
				if old_target in cur_target.parents:
					cur_target = Path(str(cur_target).replace(str(old_target), str(new_target), 1))
					break
				else:
					rename_dir_stack.pop()
					rename_dir_stack_changed = True

			when = ''
			warning = file.get('warning', '')
			target_is_dir = file.get('target_is_dir', False)
			target_is_symlink = file.get('target_is_symlink', False)
			resume = False
			try:
				if file['status'] == 'IN_PROGRESS':
					x = file.get('cur_target', None)
					if x is not None:
						cur_target = Path(x)

					if os.path.lexists(cur_target):
						resume = True

						if warning:
							if not warning.startswith('Resumed'):
								warning = f'Resumed -- {warning}'
						else:
							warning = f'Resumed'

						file['warning'] = warning
				else:
					if os.path.lexists(cur_target) and not skip_dir:
						when = 'stat_target'
						target_is_dir = cur_target.is_dir()
						target_is_symlink = cur_target.is_symlink()

						if not (file['is_dir'] and target_is_dir):
							when = 'samefile'
							if cur_file.resolve() == cur_target.resolve():
								if (mode == 'mv') or (on_conflict not in ('rename_existing', 'rename_copy')):
									raise SkippedError('Same file')

							if on_conflict == 'overwrite':
								if target_is_dir and not target_is_symlink:
									when = 'rmdir'
									os.rmdir(cur_target)
									warning = f'Overwrite'
								else:
									when = 'remove'
									os.remove(cur_target)
									warning = f'Overwrite'
							elif on_conflict == 'rename_existing':
								i = 0
								name = cur_target.name
								existing_target = cur_target
								while os.path.lexists(existing_target):
									new_name = f'{name}.rnrsave{i}'
									existing_target = existing_target.parent / new_name
									i += 1

								when = 'samefile'
								if cur_file.resolve() == cur_target.resolve():
									cur_file = existing_target

								when = 'rename'
								os.rename(cur_target, existing_target)
								warning = f'Renamed to {existing_target.name}'
							elif on_conflict == 'rename_copy':
								i = 0
								name = cur_target.name
								existing_target = cur_target
								while os.path.lexists(cur_target):
									new_name = f'{name}.rnrnew{i}'
									cur_target = cur_target.parent / new_name
									i += 1

								warning = f'Renamed to {cur_target.name}'
								if file['is_dir']:
									rename_dir_stack.append((existing_target, cur_target))
									if dbfile:
										db.set_rename_dir_stack(job_id, rename_dir_stack)
							else:
								raise SkippedError('Target exists')

					file['warning'] = warning
					file['target_is_dir'] = target_is_dir
					file['target_is_symlink'] = target_is_symlink
					file['cur_target'] = str(cur_target)

				if dbfile:
					db.update_file(file, 'IN_PROGRESS')

				if ev_abort.is_set():
					raise AbortedError()

				if ev_skip.is_set():
					ev_skip.clear()
					raise SkippedError('ev_skip')

				if rename_dir_stack_changed and dbfile:
					db.set_rename_dir_stack(job_id, rename_dir_stack)

				info['cur_source'] = str(rel_file)
				info['cur_target'] = str(cur_target)
				info['cur_size'] = file['lstat'].st_size
				info['cur_bytes'] = 0

				now = time.monotonic()
				if (now - timers['last_write']) > 0.05:
					timers['last_write'] = now
					info['cur_time'] = int(round(now - timers['cur_start']))
					info['time'] = int(round(now - timers['start']))
					q.put(info.copy())
					try:
						os.write(fd, b'\n')
					except OSError:
						pass

				if skip_dir:
					raise SkippedError('no_log')

				parent_dir = cur_target.resolve().parent

				if (mode == 'mv') and not target_is_dir:
					perform_copy = False
					try:
						os.rename(cur_file, cur_target)
						if file['is_dir']:
							skip_dir_stack.append(cur_file)
							if dbfile:
								db.set_skip_dir_stack(job_id, skip_dir_stack)
					except OSError as e:
						perform_copy = True
				else:
					perform_copy = True

				in_error = False

				if perform_copy:
					if file['is_symlink']:
						when = 'symlink'
						os.symlink(os.readlink(cur_file), cur_target)
					elif file['is_dir']:
						new_dir = False
						if not target_is_dir:
							when = 'makedirs'
							os.makedirs(cur_target, exist_ok=True)
							new_dir = True

						dir_list.append({'file': file, 'cur_file': cur_file, 'cur_target': cur_target, 'new_dir': new_dir})
						if dbfile:
							db.set_dir_list(job_id, dir_list)
					elif file['is_file']:
						when = 'copyfile'
						rnr_copyfile(cur_file, cur_target, file['lstat'].st_size, block_size, resume, info, timers, fd, q, ev_skip, ev_suspend, ev_interrupt, ev_abort)
					else:
						in_error = True
						message = f'Special file'
						error_list.append({'file': file['file'], 'message': message})
						if dbfile:
							db.set_file_status(file, 'ERROR', message)

					if not in_error:
						if not file['is_dir']:
							try:
								os.lchown(cur_target, file['lstat'].st_uid, file['lstat'].st_gid)
							except OSError as e:
								if e.errno == errno.EPERM:
									try:
										os.lchown(cur_target, -1, file['lstat'].st_gid)
									except OSError as e:
										if e.errno == errno.EPERM:
											pass
										else:
											raise
								else:
									raise

							when = 'copystat'
							shutil.copystat(cur_file, cur_target, follow_symlinks=False)

						when = 'fsync'
						parent_fd = os.open(parent_dir, 0)
						try:
							os.fsync(parent_fd)
						finally:
							os.close(parent_fd)

				if (mode == 'mv') and not file['is_dir']:
					if perform_copy:
						when = 'remove'
						os.remove(cur_file)

					when = 'fsync'
					parent_fd = os.open(cur_file.parent, 0)
					try:
						os.fsync(parent_fd)
					finally:
						os.close(parent_fd)

				if not in_error:
					completed_list.append({'file': file['file'], 'message': warning})
					if dbfile:
						db.set_file_status(file, 'DONE', warning)
			except OSError as e:
				message = f'({when}) {e.strerror} ({e.errno})'
				error_list.append({'file': file['file'], 'message': message})
				if dbfile:
					db.set_file_status(file, 'ERROR', message)
		except InterruptError as e:
			break
		except AbortedError as e:
			try:
				os.remove(cur_target)
			except OSError:
				pass

			aborted_list.extend([{'file': x['file'], 'message': ''} for x in file_list[i_file:]])
			if dbfile:
				db.set_job_status(job_id, 'ABORTED')

			break
		except SkippedError as e:
			if str(e) == 'no_log':
				if file['status'] == 'ERROR':
					error_list.append({'file': file['file'], 'message': file['message']})
				elif file['status'] == 'SKIPPED':
					skipped_list.append({'file': file['file'], 'message': file['message']})
				else:
					completed_list.append({'file': file['file'], 'message': ''})
					if dbfile and file['status'] != 'DONE':
						db.set_file_status(file, 'DONE', '')
			else:
				message = str(e)
				if message == 'ev_skip':
					message = ''
					try:
						os.remove(cur_target)
					except OSError:
						pass

				skipped_list.append({'file': file['file'], 'message': message})
				if dbfile:
					db.set_file_status(file, 'SKIPPED', message)

		total_bytes += file['lstat'].st_size
		info['bytes'] = total_bytes
		info['files'] += 1

	for entry in reversed(dir_list):
		try:
			if ev_interrupt.is_set():
				raise InterruptError()

			if dbfile and ev_nodb.is_set():
				db.delete_job(job_id)
				del db
				dbfile = None

			timers['cur_start'] = time.monotonic()

			t1 = time.monotonic()
			ev_suspend.wait()
			t2 = time.monotonic()
			dt = round(t2 - t1)
			timers['cur_start'] += dt
			timers['start'] += dt

			if ev_abort.is_set():
				raise AbortedError()

			if ev_skip.is_set():
				ev_skip.clear()
				raise SkippedError('ev_skip')

			file = entry['file']
			cur_target = entry['cur_target']
			cur_file = entry['cur_file']
			rel_file = cur_file.relative_to(cwd)
			info['cur_source'] = str(rel_file)
			info['cur_target'] = str(cur_target)
			info['cur_size'] = file['lstat'].st_size
			info['cur_bytes'] = 0

			now = time.monotonic()
			if (now - timers['last_write']) > 0.05:
				timers['last_write'] = now
				info['cur_time'] = int(round(now - timers['cur_start']))
				info['time'] = int(round(now - timers['start']))
				q.put(info.copy())
				try:
					os.write(fd, b'\n')
				except OSError:
					pass

			when = ''
			try:
				parent_dir = cur_target.resolve().parent

				if entry['new_dir']:
					try:
						os.lchown(cur_target, file['lstat'].st_uid, file['lstat'].st_gid)
					except OSError as e:
						if e.errno == errno.EPERM:
							try:
								os.lchown(cur_target, -1, file['lstat'].st_gid)
							except OSError as e:
								if e.errno == errno.EPERM:
									pass
								else:
									raise
						else:
							raise

					when = 'copystat'
					shutil.copystat(cur_file, cur_target, follow_symlinks=False)

				when = 'fsync'
				parent_fd = os.open(parent_dir, 0)
				try:
					os.fsync(parent_fd)
				finally:
					os.close(parent_fd)

				if mode == 'mv':
					when = 'rmdir'
					os.rmdir(cur_file)

					when = 'fsync'
					parent_fd = os.open(cur_file.parent, 0)
					try:
						os.fsync(parent_fd)
					finally:
						os.close(parent_fd)
			except OSError as e:
				message = f'({when}) {e.strerror} ({e.errno})'
				error_list.append({'file': file['file'], 'message': message})
				if dbfile:
					db.set_file_status(file, 'ERROR', message)

			if dbfile:
				db.set_job_status(job_id, 'DONE')
		except InterruptError as e:
			break
		except AbortedError as e:
			if dbfile:
				db.set_job_status(job_id, 'ABORTED')

			break
		except SkippedError as e:
			if str(e) == 'no_log':
				if file['status'] == 'ERROR':
					error_list.append({'file': file['file'], 'message': file['message']})
				elif file['status'] == 'SKIPPED':
					skipped_list.append({'file': file['file'], 'message': file['message']})
				else:
					completed_list.append({'file': file['file'], 'message': ''})
					if dbfile and file['status'] != 'DONE':
						db.set_file_status(file, 'DONE', '')
			else:
				message = str(e)
				if message == 'ev_skip':
					message = ''

				skipped_list.append({'file': file['file'], 'message': message})
				if dbfile:
					db.set_file_status(file, 'SKIPPED', message)

	q.put({'result': completed_list, 'error': error_list, 'skipped': skipped_list, 'aborted': aborted_list})
	try:
		os.write(fd, b'\n')
	except OSError:
		pass
	os.close(fd)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import datetime
import string
import shlex

from pathlib import Path

import urwid

from .debug_print import (debug_print, debug_pprint)


def human_readable_size(size):
	if size < 1024:
		return f'{size:d}B'

	for suffix in ['K', 'M', 'G', 'T', 'P', 'E', 'Z', 'Y']:
		size /= 1024
		if size < 1024:
			break

	return f'{size:.{max(4 - len(str(int(size))), 1)}f}{suffix}'

def format_date(d):
	d = datetime.datetime.fromtimestamp(d)
	today = datetime.date.today()
	if d.date() == today:
		return d.strftime('%H:%M').center(7)
	elif d.year == today.year:
		return d.strftime('%b %d').center(7)
	else:
		return d.strftime('%Y-%m').center(7)

def format_seconds(t):
	days, remainder = divmod(t, 86400)
	hours, remainder = divmod(remainder, 3600)
	minutes, seconds = divmod(remainder, 60)

	if days:
		return f'{days}d{hours:02d}:{minutes:02d}:{seconds:02d}'
	else:
		return f'{hours:02d}:{minutes:02d}:{seconds:02d}'

def tar_stem(file):
	p = Path(file)
	suffixes = p.suffixes
	if (len(suffixes) >= 2) and (suffixes[-2].lower() == '.tar'):
		return Path(p.stem).stem
	else:
		return p.stem

def tar_suffix(file):
	p = Path(file)
	suffixes = p.suffixes
	if (len(suffixes) >= 2) and (suffixes[-2].lower() == '.tar'):
		return ''.join(suffixes[-2:])
	else:
		return p.suffix


class TildeTextLayout(urwid.TextLayout):
	def layout(self, text, width, align, wrap):
		if len(text) <= width:
			return [[(len(text), 0, text.encode('utf-8'))]]

		full_len = max(width - 1, 2)
		half = int(full_len / 2)
		left = half
		right = full_len - left

		return [[(width, 0, f'{text[:left]}~{text[-right:]}'[:width].encode('utf-8'))]]

	def pack(self, maxcol, layout):
		maxwidth = 0
		for l in layout:
			for line in l:
				maxwidth = max(line[0], maxwidth)

		return min(maxwidth, maxcol)

TildeLayout = TildeTextLayout()


class TLineWidget(urwid.WidgetWrap):
	def __init__(self, title, title_align='center', title_attr=None, lcorner='├', tline='─', rcorner='┤'):
		self.title_widget = title
		self.title_attr = urwid.AttrMap(self.title_widget, title_attr)
		tline_divider = urwid.Divider(tline)

		if title_align == 'left':
			tline_widgets = [('pack', self.title_attr), tline_divider]
		else:
			tline_widgets = [tline_divider, ('pack', self.title_attr)]
			if title_align == 'center':
				tline_widgets.append(tline_divider)

		self.tline_widget = urwid.Columns(tline_widgets)

		w = urwid.Columns([(1, urwid.Text(lcorner)), (1, urwid.Text(tline)), self.tline_widget, (1, urwid.Text(tline)), (1, urwid.Text(rcorner))])

		super().__init__(w)

	def set_title(self, text):
		self.title_widget.set_text(text)

	def set_title_attr(self, attr):
		self.title_attr.set_attr_map({None: attr})


class InterruptError(Exception):
	pass

class AbortedError(Exception):
	pass

class SkippedError(Exception):
	pass


class Template(string.Template):
	delimiter = '%'

def apply_template(text, screen, quote=True):
	if quote:
		fn_quote = shlex.quote
	else:
		fn_quote = str

	cwd = str(screen.center.focus.cwd)

	try:
		current_file = fn_quote(str(screen.center.focus.get_focus()['file'].relative_to(cwd)))
		current_name = fn_quote(tar_stem(screen.center.focus.get_focus()['file']))
		current_extension = fn_quote(tar_suffix(screen.center.focus.get_focus()['file']))
	except (TypeError, AttributeError):
		current_file = fn_quote('')
		current_name = fn_quote('')
		current_extension = fn_quote('')

	current_tagged = ' '.join([fn_quote(str(x.relative_to(cwd))) for x in screen.center.focus.get_tagged_files()])
	if not current_tagged:
		current_tagged = fn_quote('')

	if screen.center.focus == screen.left:
		other = screen.right
	else:
		other = screen.left

	other_cwd = str(other.cwd)

	try:
		other_file = fn_quote(str(other.get_focus()['file']))
		other_name = fn_quote(tar_stem(other.get_focus()['file']))
		other_extension = fn_quote(tar_suffix(other.get_focus()['file']))
	except (TypeError, AttributeError):
		other_file = fn_quote('')
		other_name = fn_quote('')
		other_extension = fn_quote('')

	other_tagged = ' '.join([fn_quote(str(x)) for x in other.get_tagged_files()])
	if not current_tagged:
		other_tagged = fn_quote('')

	s = Template(text)
	d = {
		'f': current_file,
		'n': current_name,
		'e': current_extension,
		'd': fn_quote(cwd),
		'b': fn_quote(Path(cwd).name),
		's': current_tagged,
		't': current_tagged,
		'F': other_file,
		'N': other_name,
		'E': other_extension,
		'D': fn_quote(other_cwd),
		'B': fn_quote(Path(other_cwd).name),
		'S': other_tagged,
		'T': other_tagged,
	}

	return s.safe_substitute(d)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

from .import_config import *


PALETTE = [
	('default', 'default', 'default'),
	('default_error', 'light red', 'default', 'default,bold'),

	('panel', PANEL_FG, PANEL_BG),
	('reverse', REVERSE_FG, REVERSE_BG, 'standout'),
	('selected', SELECTED_FG, SELECTED_BG, 'standout'),
	('marked', MARKED_FG, PANEL_BG, 'default,bold'),
	('markselect', MARKSELECT_FG, SELECTED_BG, 'standout,bold'),

	('directory', DIRECTORY_FG, PANEL_BG),
	('dir_symlink', DIR_SYMLINK_FG, PANEL_BG),
	('executable', EXECUTABLE_FG, PANEL_BG),
	('symlink', SYMLINK_FG, PANEL_BG),
	('stalelink', STALELINK_FG, PANEL_BG),
	('device', DEVICE_FG, PANEL_BG),
	('special', SPECIAL_FG, PANEL_BG),
	('archive', ARCHIVE_FG, PANEL_BG),

	('hotkey', HOTKEY_FG, HOTKEY_BG),

	('error', ERROR_FG, ERROR_BG, 'standout'),
	('error_title', ERROR_TITLE_FG, ERROR_BG, 'standout'),
	('error_focus', ERROR_FOCUS_FG, ERROR_FOCUS_BG, 'standout'),

	('dialog', DIALOG_FG, DIALOG_BG, 'standout'),
	('dialog_title', DIALOG_TITLE_FG, DIALOG_BG, 'standout'),
	('dialog_focus', DIALOG_FOCUS_FG, DIALOG_FOCUS_BG, 'standout'),
	('progress', DIALOG_BG, DIALOG_FG),
	('input', INPUT_FG, INPUT_BG),

	('Text', TEXT_FG, TEXT_BG),
	('Namespace', NAMESPACE_FG, TEXT_BG),
	('Keyword', KEYWORD_FG, TEXT_BG),
	('Class', CLASS_FG, TEXT_BG),
	('Operator', OPERATOR_FG, TEXT_BG),
	('String', STRING_FG, TEXT_BG),
	('Literal', LITERAL_FG, TEXT_BG),
	('Comment', COMMENT_FG, TEXT_BG),
	('Lineno', LINENO_FG, TEXT_BG, 'default,bold'),
]




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import urwid

from .debug_print import (debug_print, debug_pprint)


class DlgError(urwid.WidgetWrap):
	def __init__(self, screen, e, title='Error', error=True):
		self.screen = screen

		if error:
			attr = 'error'
			title_attr = 'error_title'
		else:
			attr = 'dialog'
			title_attr = 'dialog_title'

		w = urwid.Filler(urwid.Text(f' {e} ', align='center'), top=1, bottom=1)
		w = urwid.LineBox(w, title, title_attr=title_attr)
		w = urwid.Padding(w, left=1, right=1)
		w = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			(5, w),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w = urwid.AttrMap(w, attr)

		super().__init__(w)

	def keypress(self, size, key):
		self.screen.close_dialog()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import argparse
import shutil
import stat
import signal
import functools

from pathlib import Path
from queue import Queue
from threading import (Thread, Event)

import urwid

import xdg.BaseDirectory

from . import __version__
from . import rnrview

from .import_config import *
from .palette import PALETTE
from .panel import Panel
from .preview_panel import PreviewPanel
from .cmdbar import CmdBar
from .buttonbar import ButtonBar
from .bookmarks import (Bookmarks, BOOKMARK_KEYS)
from .dlg_error import DlgError
from .dlg_question import DlgQuestion
from .dlg_dirscan import DlgDirscan
from .rnr_dirscan import rnr_dirscan
from .dlg_delete_progress import DlgDeleteProgress
from .rnr_delete import rnr_delete
from .dlg_report import DlgReport
from .dlg_cpmv import DlgCpMv
from .dlg_cpmv_progress import DlgCpMvProgress
from .rnr_cpmv import rnr_cpmv
from .database import DataBase
from .dlg_pending_job import DlgPendingJob
from .debug_print import (debug_print, debug_pprint, set_debug_fh)


DATA_DIR = Path(xdg.BaseDirectory.save_data_path('rnr'))


Labels = [
	' ', #'Help',
	' ', #'Menu',
	'View',
	'Edit',
	'Copy',
	'Move',
	'Mkdir',
	'Delete',
	' ', #'PullDn',
	'Quit',
]


class Screen(urwid.WidgetWrap):
	def __init__(self, controller):
		self.left = Panel(controller)
		self.right = Panel(controller)
		self.preview_panel = PreviewPanel(controller)
		#self.center = urwid.Columns([self.left, self.preview_panel, self.right])
		self.center = urwid.Columns([self.left, self.right])
		self.command_bar = CmdBar(controller, self)
		w = urwid.Filler(self.command_bar)
		pile_widgets = [self.center, (1, w)]

		self.bottom = ButtonBar(Labels)
		w = urwid.Filler(self.bottom)
		if SHOW_BUTTONBAR:
			pile_widgets.append((1, w))

		self.pile = urwid.Pile(pile_widgets)
		self.pile.focus_position = 0
		self.main_area = 0

		self.list_box = self.preview_panel.listbox

		self.show_preview = False
		self.in_error = False

		super().__init__(self.pile)

	def update_focus(self):
		for i, e in enumerate(self.center.contents):
			if i == self.center.focus_position:
				e[0].set_title_attr('reverse')
				e[0].focused = True
				try:
					e[0].show_preview(e[0].get_focus())
				except AttributeError:
					pass
			else:
				e[0].set_title_attr('panel')
				e[0].focused = False

	def close_dialog(self):
		self.pile.contents[self.main_area] = (self.center, self.pile.options())

		try:
			self.center.focus.remove_force_focus()
		except AttributeError:
			pass

		self.in_error = False

	def error(self, e, title='Error', error=True):
		try:
			self.center.focus.force_focus()
		except AttributeError:
			pass

		self.pile.contents[self.main_area] = (urwid.Overlay(DlgError(self, e, title, error), self.center,
			'center', len(e) + 6,
			'middle', 'pack',
		), self.pile.options())

		self.in_error = True


class App(object):
	def __init__(self, printwd, dbfile, monochrome, tabsize):
		self.printwd = printwd

		if not dbfile:
			self.dbfile = None
		elif dbfile == ':memory:':
			self.dbfile = dbfile
		else:
			self.dbfile = str(Path(dbfile).resolve())

		self.monochrome = monochrome
		self.tabsize = tabsize

		if self.dbfile:
			db = DataBase(self.dbfile)
			del db

		self.opener = OPENER
		self.pager = PAGER
		self.editor = EDITOR
		self.use_internal_viewer = USE_INTERNAL_VIEWER

		self.old_screen = None
		self.loop = None
		self.screen = Screen(self)
		self.screen.update_focus()
		self.leader = ''
		self.ev_interrupt = Event()
		self.suspend = set()
		self.pending_jobs = []
		self.focused_quickviewer = False

		self.bookmarks = Bookmarks(CONFIG_DIR / 'bookmarks')
		if 'h' not in self.bookmarks:
			self.bookmarks['h'] = Path.home()

		self.check_pending_jobs()

	def run(self):
		self.loop = urwid.MainLoop(self.screen, PALETTE, unhandled_input=self.keypress)

		if self.monochrome:
			self.loop.screen.set_terminal_properties(colors=1)

		signal.signal(signal.SIGINT, self.signal_handler)
		signal.signal(signal.SIGTERM, self.signal_handler)

		self.loop.run()

	def signal_handler(self, signum, frame):
		try:
			self.loop.stop()
		except AttributeError:
			pass

		if signum == signal.SIGINT:
			print('Ctrl+C')
		elif signum == signal.SIGTERM:
			print('Kill')

		self.ev_interrupt.set()

		for ev in self.suspend:
			ev.set()

		sys.exit(1)

	def keypress(self, key):
		if key == 'esc':
			self.screen.command_bar.reset()
			self.leader = ''
		elif self.leader == 's':
			if key == 'n':
				self.screen.left.sort('sort_by_name')
				self.screen.right.sort('sort_by_name')
			elif key == 'N':
				self.screen.left.sort('sort_by_name', reverse=True)
				self.screen.right.sort('sort_by_name', reverse=True)
			elif key == 'e':
				self.screen.left.sort('sort_by_extension')
				self.screen.right.sort('sort_by_extension')
			elif key == 'E':
				self.screen.left.sort('sort_by_extension', reverse=True)
				self.screen.right.sort('sort_by_extension', reverse=True)
			elif key == 'd':
				self.screen.left.sort('sort_by_date')
				self.screen.right.sort('sort_by_date')
			elif key == 'D':
				self.screen.left.sort('sort_by_date', reverse=True)
				self.screen.right.sort('sort_by_date', reverse=True)
			elif key == 's':
				self.screen.left.sort('sort_by_size')
				self.screen.right.sort('sort_by_size')
			elif key == 'S':
				self.screen.left.sort('sort_by_size', reverse=True)
				self.screen.right.sort('sort_by_size', reverse=True)

			self.screen.command_bar.reset()
			self.leader = ''
		elif self.leader == 'm':
			if key in BOOKMARK_KEYS:
				self.bookmarks[key] = self.screen.center.focus.cwd

			self.screen.command_bar.reset()
			self.leader = ''
		elif self.leader in ('`', "'"):
			if key in ('`', "'"):
				if self.screen.center.focus.old_cwd != self.screen.center.focus.cwd:
					self.screen.center.focus.chdir(self.screen.center.focus.old_cwd)
			elif key in BOOKMARK_KEYS:
				try:
					if self.bookmarks[key] != str(self.screen.center.focus.cwd):
						self.screen.center.focus.chdir(Path(self.bookmarks[key]))
				except KeyError:
					pass

			self.screen.command_bar.reset()
			self.leader = ''
		elif self.leader == 'u':
			if key == 'v':
				self.screen.center.focus.untag_all()
			elif key in ('f', '/'):
				self.screen.center.focus.filter('')

			self.screen.command_bar.reset()
			self.leader = ''
		elif self.leader == 'c':
			self.screen.command_bar.reset()
			self.leader = ''
			if key in ('c', 'w'):
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='replace')
				except TypeError as e:
					pass
			elif key == 'e':
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='replace_before')
				except TypeError:
					pass
		else:
			if key in ('q', 'Q', 'f10'):
				self.quit()
			elif key == 'tab':
				if self.screen.pile.focus_position == 0:
					self.screen.center.focus_position = (self.screen.center.focus_position + 1) % len(self.screen.center.contents)
					self.screen.update_focus()
					if self.screen.center.focus == self.screen.preview_panel:
						self.focused_quickviewer = True
						self.set_input_rnrview()
					else:
						self.focused_quickviewer = False
			elif key == 'shift tab':
				if self.screen.pile.focus_position == 0:
					self.screen.center.focus_position = (self.screen.center.focus_position - 1) % len(self.screen.center.contents)
					self.screen.update_focus()
					if self.screen.center.focus == self.screen.preview_panel:
						self.focused_quickviewer = True
						self.set_input_rnrview()
					else:
						self.focused_quickviewer = False
			elif key in ('f', '/'):
				self.screen.command_bar.filter()
			elif key == 'enter':
				self.screen.command_bar.execute()
			elif key == 'backspace':
				self.screen.left.toggle_hidden()
				self.screen.right.toggle_hidden()
			elif key == 's':
				self.leader = key
				self.screen.command_bar.set_leader(self.leader)
			elif key == 'm':
				self.leader = key
				self.screen.command_bar.set_leader(self.leader)
			elif key in ('`', "'"):
				self.leader = key
				self.screen.command_bar.set_leader(self.leader)
			elif key == 'u':
				self.leader = key
				self.screen.command_bar.set_leader(self.leader)
			elif key == 'meta i':
				cwd = self.screen.center.focus.cwd

				if (self.screen.left is not self.screen.center.focus) and (self.screen.left.cwd != cwd):
					self.screen.left.chdir(cwd)

				if (self.screen.right is not self.screen.center.focus) and (self.screen.right.cwd != cwd):
					self.screen.right.chdir(cwd)
			elif key == 'meta o':
				cwd = self.screen.center.focus.cwd.parent
				obj = self.screen.center.focus.get_focus()
				try:
					if stat.S_ISDIR(obj['stat'].st_mode):
						cwd = obj['file']
				except TypeError:
					pass

				if (self.screen.left is not self.screen.center.focus) and (self.screen.left.cwd != cwd):
					self.screen.left.chdir(cwd)

				if (self.screen.right is not self.screen.center.focus) and (self.screen.right.cwd != cwd):
					self.screen.right.chdir(cwd)
			elif key == 'ctrl r':
				self.reload()
			elif key == 'ctrl u':
				(self.screen.center.contents[0], self.screen.center.contents[1]) = (self.screen.center.contents[1], self.screen.center.contents[0])
				self.screen.center.focus_position ^= 1
				self.screen.update_focus()
			elif key == 'ctrl o':
				self.loop.stop()
				input('Press ENTER to continue...')
				self.loop.start()
				os.kill(os.getpid(), signal.SIGWINCH)
				self.reload()
			elif key == 'ctrl q':
				pos = self.screen.center.focus_position ^ 1
				if self.screen.left is self.screen.center.focus:
					if self.screen.show_preview:
						self.screen.show_preview = False
						self.screen.center.contents[pos] = (self.screen.right, self.screen.center.options())
					else:
						self.screen.show_preview = True
						self.screen.center.contents[pos] = (self.screen.preview_panel, self.screen.center.options())
				elif self.screen.right is self.screen.center.focus:
					if self.screen.show_preview:
						self.screen.show_preview = False
						self.screen.center.contents[pos] = (self.screen.left, self.screen.center.options())
					else:
						self.screen.show_preview = True
						self.screen.center.contents[pos] = (self.screen.preview_panel, self.screen.center.options())
				else:
					if self.screen.center.contents[pos][0] == self.screen.left:
						self.screen.show_preview = False
						self.screen.center.contents[self.screen.center.focus_position] = (self.screen.right, self.screen.center.options())
					else:
						self.screen.show_preview = False
						self.screen.center.contents[self.screen.center.focus_position] = (self.screen.left, self.screen.center.options())


				self.screen.update_focus()
				self.reload()
			elif key == 'f7':
				self.screen.command_bar.mkdir(self.screen.center.focus.cwd)
			elif key == 'c':
				self.leader = key
				self.screen.command_bar.set_leader(self.leader)
			elif key == 'r':
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='replace')
				except TypeError:
					pass
			elif key in ('i', 'I'):
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='insert')
				except TypeError:
					pass
			elif key == 'a':
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='append_before')
				except TypeError:
					pass
			elif key == 'A':
				obj = self.screen.center.focus.get_focus()
				try:
					self.screen.command_bar.rename(obj['file'], mode='append_after')
				except TypeError:
					pass
			elif key == '+':
				self.screen.command_bar.tag_glob()
			elif key in ('-', '\\'):
				self.screen.command_bar.untag_glob()
			elif key in (':', '!'):
				self.screen.command_bar.shell()
			elif key == 'f8':
				tagged_files = self.screen.center.focus.get_tagged_files()
				if tagged_files:
					if len(tagged_files) == 1:
						question = f'Delete {tagged_files[0].name}?'
					else:
						question = f'Delete {len(tagged_files)} files/directories?'

					self.screen.center.focus.force_focus()
					self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(DlgQuestion(self, title='Delete', question=question,
						on_yes=lambda x: self.on_delete(tagged_files, str(self.screen.center.focus.cwd)), on_no=lambda x: self.screen.close_dialog()), self.screen.center,
						'center', max(len(question) + 6, 21),
						'middle', 'pack',
					), self.screen.pile.options())
			elif key == 'f5':
				tagged_files = self.screen.center.focus.get_tagged_files()
				if tagged_files:
					if len(tagged_files) == 1:
						question = f'Copy {tagged_files[0].name} to:'
					else:
						question = f'Copy {len(tagged_files)} files/directories to:'

					if self.screen.center.focus == self.screen.left:
						dest_dir = self.screen.right.cwd
					else:
						dest_dir = self.screen.left.cwd

					self.screen.center.focus.force_focus()
					self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(DlgCpMv(self, title='Copy', question=question, dest_dir=str(dest_dir),
						on_ok=functools.partial(self.on_copy, tagged_files, str(self.screen.center.focus.cwd)), on_cancel=lambda x: self.screen.close_dialog()), self.screen.center,
						'center', ('relative', 85),
						'middle', 'pack',
					), self.screen.pile.options())
			elif key == 'f6':
				tagged_files = self.screen.center.focus.get_tagged_files()
				if tagged_files:
					if len(tagged_files) == 1:
						question = f'Move {tagged_files[0].name} to:'
					else:
						question = f'Move {len(tagged_files)} files/directories to:'

					if self.screen.center.focus == self.screen.left:
						dest_dir = self.screen.right.cwd
					else:
						dest_dir = self.screen.left.cwd

					self.screen.center.focus.force_focus()
					self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(DlgCpMv(self, title='Move', question=question, dest_dir=str(dest_dir),
						on_ok=functools.partial(self.on_move, tagged_files, str(self.screen.center.focus.cwd)), on_cancel=lambda x: self.screen.close_dialog()), self.screen.center,
						'center', ('relative', 85),
						'middle', 'pack',
					), self.screen.pile.options())

	def reload(self, focus_path=None, old_focus=None, only_focused=False):
		if old_focus is None:
			if (not only_focused) or (self.screen.center.focus == self.screen.left):
				left_path = focus_path
			else:
				left_path = None

			if (not only_focused) or (self.screen.center.focus == self.screen.right):
				right_path = focus_path
			else:
				right_path = None
		else:
			obj = self.screen.left.get_focus()
			left_path = None
			try:
				if obj['file'] == old_focus:
					left_path = focus_path
			except TypeError:
				pass

			obj = self.screen.right.get_focus()
			right_path = None
			try:
				if obj['file'] == old_focus:
					right_path = focus_path
			except TypeError:
				pass

		self.screen.left.reload(left_path)
		self.screen.right.reload(right_path)

	def do_dirscan(self, files, cwd, on_complete):
		self.screen.center.focus.force_focus()

		q = Queue()
		ev_abort = Event()
		ev_skip = Event()
		dlg = DlgDirscan(self, cwd, q, ev_abort, ev_skip, on_complete)
		self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
			'center', ('relative', 50),
			'middle', 'pack',
		), self.screen.pile.options())

		fd = self.loop.watch_pipe(dlg.on_pipe_data)
		dlg.fd = fd

		Thread(target=rnr_dirscan, args=(files, cwd, fd, q, self.ev_interrupt, ev_abort, ev_skip)).start()

	def on_finish(self, completed_list, error_list, skipped_list, aborted_list, operation, files, cwd, dest, scan_error, scan_skipped, job_id):
		warnings = [x for x in completed_list if x['message']]
		if scan_error or error_list or scan_skipped or skipped_list or aborted_list or warnings:
			self.screen.center.focus.force_focus()

			dlg = DlgReport(self, completed_list, error_list, skipped_list, aborted_list, operation, files, cwd, dest, scan_error, scan_skipped, job_id)
			self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
				'center', ('relative', 75),
				'middle', ('relative', 75),
			), self.screen.pile.options())
		else:
			if self.dbfile:
				db = DataBase(self.dbfile)
				db.delete_job(job_id)
				del db

			self.reload()

	def on_delete(self, files, cwd):
		self.screen.close_dialog()
		self.do_dirscan(files, cwd, functools.partial(self.do_delete, files=files, cwd=cwd, job_id=None))

	def do_delete(self, file_list, scan_error, scan_skipped, files, cwd, job_id):
		self.screen.center.focus.force_focus()

		if self.dbfile and (job_id is None):
			db = DataBase(self.dbfile)
			job_id = db.new_job('Delete', file_list, scan_error, scan_skipped, files, cwd)
			del db

		q = Queue()
		ev_skip = Event()
		ev_suspend = Event()
		ev_suspend.set()
		self.suspend.add(ev_suspend)
		ev_abort = Event()
		ev_nodb = Event()
		dlg = DlgDeleteProgress(self, len(file_list), sum((x['lstat'].st_size for x in file_list)), q, ev_skip, ev_suspend, ev_abort, ev_nodb, functools.partial(self.on_finish, operation='Delete', files=files, cwd=cwd, dest=None, scan_error=scan_error, scan_skipped=scan_skipped, job_id=job_id))
		self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
			'center', ('relative', 75),
			'middle', 'pack',
		), self.screen.pile.options())

		fd = self.loop.watch_pipe(dlg.on_pipe_data)
		dlg.fd = fd

		Thread(target=rnr_delete, args=(file_list, fd, q, ev_skip, ev_suspend, self.ev_interrupt, ev_abort, ev_nodb, self.dbfile, job_id)).start()

	def on_copy(self, files, cwd, dest, on_conflict):
		self.screen.close_dialog()

		path_cwd = Path(cwd)
		path_dest = Path(dest)
		try:
			path_dest = path_dest.expanduser()
		except RuntimeError:
			pass

		if not path_dest.is_absolute():
			path_dest = Path(os.path.normpath(path_cwd / path_dest))

		try:
			if len(files) == 1:
				if path_dest.is_dir():
					if (path_cwd.resolve() == path_dest.resolve()) and (on_conflict in ('overwrite', 'skip')):
						pass
					else:
						self.do_dirscan(files, cwd, functools.partial(self.do_copy, files=files, cwd=cwd, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
				else:
					dest_parent = path_dest.parent
					if not dest_parent.is_dir():
						self.screen.error(f'{str(Path(dest).parent)} is not a directory')
					elif (path_cwd.resolve() == path_dest.resolve()) and (on_conflict in ('overwrite', 'skip')):
						pass
					else:
						self.do_dirscan(files, cwd, functools.partial(self.do_copy, files=files, cwd=cwd, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
			else:
				if not path_dest.is_dir():
					self.screen.error(f'{dest} is not a directory')
				elif (path_cwd.resolve() == path_dest.resolve()) and (on_conflict in ('overwrite', 'skip')):
					pass
				else:
					self.do_dirscan(files, cwd, functools.partial(self.do_copy, cwd=cwd, files=files, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
		except (FileNotFoundError, PermissionError) as e:
			self.screen.error(f'{e.strerror} ({e.errno})')

	def do_copy(self, file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict, job_id):
		self.screen.center.focus.force_focus()

		if self.dbfile and (job_id is None):
			db = DataBase(self.dbfile)
			job_id = db.new_job('Copy', file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict)
			del db

		q = Queue()
		ev_skip = Event()
		ev_suspend = Event()
		ev_suspend.set()
		self.suspend.add(ev_suspend)
		ev_abort = Event()
		ev_nodb = Event()
		dlg = DlgCpMvProgress(self, 'Copy', len(file_list), sum((x['lstat'].st_size for x in file_list)), q, ev_skip, ev_suspend, ev_abort, ev_nodb, functools.partial(self.on_finish, operation='Copy', files=files, cwd=cwd, dest=dest, scan_error=scan_error, scan_skipped=scan_skipped, job_id=job_id))
		self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
			'center', ('relative', 75),
			'middle', 'pack',
		), self.screen.pile.options())

		fd = self.loop.watch_pipe(dlg.on_pipe_data)
		dlg.fd = fd

		Thread(target=rnr_cpmv, args=('cp', file_list, cwd, dest, on_conflict, fd, q, ev_skip, ev_suspend, self.ev_interrupt, ev_abort, ev_nodb, self.dbfile, job_id)).start()

	def on_move(self, files, cwd, dest, on_conflict):
		self.screen.close_dialog()

		path_cwd = Path(cwd)
		path_dest = Path(dest)
		try:
			path_dest = path_dest.expanduser()
		except RuntimeError:
			pass

		if not path_dest.is_absolute():
			path_dest = Path(os.path.normpath(path_cwd / path_dest))

		try:
			if len(files) == 1:
				if path_dest.is_dir():
					if path_cwd.resolve() == path_dest.resolve():
						pass
					else:
						self.do_dirscan(files, cwd, functools.partial(self.do_move, files=files, cwd=cwd, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
				else:
					dest_parent = path_dest.parent
					if not dest_parent.is_dir():
						self.screen.error(f'{str(Path(dest).parent)} is not a directory')
					elif path_cwd.resolve() == path_dest.resolve():
						pass
					else:
						self.do_dirscan(files, cwd, functools.partial(self.do_move, files=files, cwd=cwd, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
			else:
				if not path_dest.is_dir():
					self.screen.error(f'{dest} is not a directory')
				elif path_cwd.resolve() == path_dest.resolve():
					pass
				else:
					self.do_dirscan(files, cwd, functools.partial(self.do_move, files=files, cwd=cwd, dest=str(path_dest), on_conflict=on_conflict, job_id=None))
		except (FileNotFoundError, PermissionError) as e:
			self.screen.error(f'{e.strerror} ({e.errno})')

	def do_move(self, file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict, job_id):
		self.screen.center.focus.force_focus()

		if self.dbfile and (job_id is None):
			db = DataBase(self.dbfile)
			job_id = db.new_job('Move', file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict)
			del db

		q = Queue()
		ev_skip = Event()
		ev_suspend = Event()
		ev_suspend.set()
		self.suspend.add(ev_suspend)
		ev_abort = Event()
		ev_nodb = Event()
		dlg = DlgCpMvProgress(self, 'Move', len(file_list), sum((x['lstat'].st_size for x in file_list)), q, ev_skip, ev_suspend, ev_abort, ev_nodb, functools.partial(self.on_finish, operation='Move', files=files, cwd=cwd, dest=dest, scan_error=scan_error, scan_skipped=scan_skipped, job_id=job_id))
		self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
			'center', ('relative', 75),
			'middle', 'pack',
		), self.screen.pile.options())

		fd = self.loop.watch_pipe(dlg.on_pipe_data)
		dlg.fd = fd

		Thread(target=rnr_cpmv, args=('mv', file_list, cwd, dest, on_conflict, fd, q, ev_skip, ev_suspend, self.ev_interrupt, ev_abort, ev_nodb, self.dbfile, job_id)).start()

	def check_pending_jobs(self):
		if not self.dbfile:
			return

		db = DataBase(self.dbfile)
		self.pending_jobs.extend(db.get_jobs())
		del db

		if self.pending_jobs:
			self.show_next_pending_job()

	def show_next_pending_job(self):
		self.screen.center.focus.force_focus()

		pending_job = self.pending_jobs.pop(0)
		dlg = DlgPendingJob(self, pending_job)
		self.screen.pile.contents[self.screen.main_area] = (urwid.Overlay(dlg, self.screen.center,
			'center', ('relative', 75),
			'middle', ('relative', 75),
		), self.screen.pile.options())

	def view(self, filename):
		try:
			file_size = os.stat(filename).st_size
			screen = rnrview.Screen(self, filename, file_size, self.tabsize)
		except OSError:
			return

		self.old_screen = self.screen
		self.screen = screen
		self.loop.widget = self.screen
		self.set_input_rnrview()

	def close_viewer(self, key):
		if self.old_screen:
			self.screen = self.old_screen
			self.old_screen = None
			self.loop.widget = self.screen
			self.set_input_rnr()
			self.reload()
		elif key in ('q', 'Q', 'f10'):
			self.quit()

	def set_input_rnrview(self):
		self.screen.bottom.set_labels(rnrview.Labels)
		self.loop._unhandled_input = functools.partial(rnrview.keypress, self)

	def set_input_rnr(self):
		self.screen.bottom.set_labels(Labels)
		self.loop._unhandled_input = self.keypress

	def quit(self):
		if self.printwd:
			try:
				with open(self.printwd, 'w') as fh:
					fh.write(str(self.screen.center.focus.cwd))
			except (FileNotFoundError, PermissionError):
				pass

		raise urwid.ExitMainLoop()


def main():
	parser = argparse.ArgumentParser()
	parser.add_argument('-V', '--version', action='version', version=f'%(prog)s {__version__}')
	parser.add_argument('-P', '--printwd', help='Print last working directory to specified file', metavar='<file>')
	parser.add_argument('-D', '--database', help='Specify database file to use (default: %(default)s)', metavar='<file>', default=str(DATA_DIR / 'rnr.db'), dest='dbfile')
	parser.add_argument('-n', '--nodb', help='Do not use database', action='store_false', dest='use_db')
	parser.add_argument('-b', '--nocolor', help='Requests to run in black and white', action='store_true', dest='monochrome')
	parser.add_argument('-t', '--tabsize', help='set tab size for viewer (default: %(default)d)', type=int, default=TAB_SIZE)
	parser.add_argument('-d', '--debug', help='activate debug mode', action='store_true')
	args = parser.parse_args()

	if args.debug:
		set_debug_fh(open(Path.home() / 'rnr.log', 'w', buffering=1))

	if args.use_db:
		dbfile = args.dbfile
	else:
		dbfile = None

	app = App(args.printwd, dbfile, args.monochrome, args.tabsize)
	app.run()


if __name__ == '__main__':
	sys.exit(main())




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import json

from pathlib import Path

import urwid

from atomicwrites import atomic_write

from .database import DataBase
from .utils import (human_readable_size, format_seconds, TildeLayout)
from .debug_print import (debug_print, debug_pprint)


class DlgPendingJob(urwid.WidgetWrap):
	def __init__(self, controller, pending_job):
		self.controller = controller
		self.command_bar = controller.screen.command_bar
		self.pending_job = pending_job

		attr = 'dialog'
		title_attr = 'dialog_title'
		focus_attr = 'dialog_focus'

		self.messages = [
			f'Status: {pending_job["status"]}',
			f'Operation: {pending_job["operation"]}',
			f'From: {str(pending_job["cwd"])}',
		]

		if pending_job['dest']:
			self.messages.append(f'To: {str(pending_job["dest"])}')

		self.messages.append('Files:')
		for file in json.loads(pending_job['files']):
			self.messages.append(f'{str(Path(file).relative_to(pending_job["cwd"]))}')

		l = [urwid.Text(x, layout=TildeLayout) for x in self.messages]
		w = urwid.SimpleListWalker(l)
		self.listbox = urwid.ListBox(w)
		w = urwid.LineBox(urwid.Padding(self.listbox, left=1, right=1), 'Interrupted Job', title_attr=title_attr, bline='')
		top = urwid.Padding(w, left=1, right=1)

		self.btn_continue = urwid.Button('Continue', lambda x: self.on_continue())
		attr_btn_continue = urwid.AttrMap(self.btn_continue, attr, focus_attr)
		self.btn_skip = urwid.Button('Skip', lambda x: self.on_skip())
		attr_btn_skip = urwid.AttrMap(self.btn_skip, attr, focus_attr)
		self.btn_abort = urwid.Button('Abort', lambda x: self.on_abort())
		attr_btn_abort = urwid.AttrMap(self.btn_abort, attr, focus_attr)
		w = urwid.Columns([urwid.Divider(' '), (12, attr_btn_continue), (1, urwid.Text(' ')), (8, attr_btn_skip), (1, urwid.Text(' ')), (9, attr_btn_abort), urwid.Divider(' ')])
		w = urwid.LineBox(urwid.Filler(w), tlcorner='├', trcorner='┤')
		bottom = urwid.Padding(w, left=1, right=1)

		w = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			top,
			(3, bottom),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w.set_focus(2)
		w = urwid.AttrMap(w, attr)

		super().__init__(w)

	def keypress(self, size, key):
		if key in ('left', 'right', ' ', 'enter'):
			return super().keypress(size, key)
		elif key == 'h':
			return super().keypress(size, 'left')
		elif key == 'l':
			return super().keypress(size, 'right')
		elif key in ('j', 'down', 'ctrl f', 'page down'):
			self.listbox.keypress(size, 'page down')
		elif key in ('k', 'up', 'ctrl b', 'page up'):
			self.listbox.keypress(size, 'page up')
		elif key in ('g', 'home'):
			self.listbox.keypress(size, 'home')
		elif key in ('G', 'end'):
			self.listbox.keypress(size, 'end')

	def on_continue(self):
		self.controller.screen.close_dialog()

		job_id = self.pending_job['id']

		if self.controller.dbfile:
			db = DataBase(self.controller.dbfile)
			file_list = db.get_file_list(job_id)
			del db
		else:
			file_list = []

		scan_error = json.loads(self.pending_job['scan_error'])
		scan_skipped = json.loads(self.pending_job['scan_skipped'])
		files = json.loads(self.pending_job['files'])
		cwd = self.pending_job['cwd']
		dest = self.pending_job['dest']
		on_conflict = self.pending_job['on_conflict']
		operation = self.pending_job['operation']

		if self.pending_job['status'] in ('ABORTED', 'DONE'):
			if operation == 'Delete':
				file_list.sort(key=lambda x: x['file'], reverse=True)
			else:
				file_list.sort(key=lambda x: x['file'])

			error_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'ERROR']
			skipped_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'SKIPPED']
			completed_list = [{'file': x['file'], 'message': x['message']} for x in file_list if x['status'] == 'DONE']
			if self.pending_job['status'] == 'ABORTED':
				aborted_list = [{'file': x['file'], 'message': (x['message'] or '')} for x in file_list if x['status'] not in ('DONE', 'ERROR', 'SKIPPED')]
			else:
				aborted_list = []

			self.controller.on_finish(completed_list, error_list, skipped_list, aborted_list, operation, files, cwd, dest, scan_error, scan_skipped, job_id)
		else:
			if operation == 'Delete':
				self.controller.do_delete(file_list, scan_error, scan_skipped, files, cwd, job_id)
			elif operation == 'Copy':
				self.controller.do_copy(file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict, job_id)
			elif operation == 'Move':
				self.controller.do_move(file_list, scan_error, scan_skipped, files, cwd, dest, on_conflict, job_id)
			else:
				if self.controller.pending_jobs:
					self.controller.show_next_pending_job()
				else:
					self.controller.reload()

	def on_skip(self):
		self.controller.screen.close_dialog()

		if self.controller.pending_jobs:
			self.controller.show_next_pending_job()
		else:
			self.controller.reload()

	def on_abort(self):
		if self.controller.dbfile:
			db = DataBase(self.controller.dbfile)
			db.delete_job(self.pending_job['id'])
			del db

		self.controller.screen.close_dialog()

		if self.controller.pending_jobs:
			self.controller.show_next_pending_job()
		else:
			self.controller.reload()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import urwid

from .utils import (human_readable_size, format_seconds, TildeLayout, apply_template)


class DlgCpMv(urwid.WidgetWrap):
	def __init__(self, controller, title, question, dest_dir, on_ok, on_cancel=None):
		self.controller = controller
		self.on_ok = on_ok

		label = urwid.Text(question, layout=TildeLayout)
		self.edit = urwid.Edit(edit_text=dest_dir.replace('%', '%%'), wrap='clip')
		w = urwid.AttrMap(self.edit, 'input', 'input')
		w = urwid.SimpleFocusListWalker([
			label,
			w,
		])
		w = urwid.ListBox(w)
		w = urwid.LineBox(urwid.Padding(w, left=1, right=1), title, title_attr='dialog_title', bline='')
		top = urwid.Padding(w, left=1, right=1)

		label = urwid.Text('On conflict:')
		bgroup = []
		self.btn_overwrite = urwid.RadioButton(bgroup, 'Overwrite')
		attr_btn_overwrite = urwid.AttrMap(self.btn_overwrite, 'dialog', 'dialog_focus')
		self.btn_skip = urwid.RadioButton(bgroup, 'Skip')
		attr_btn_skip = urwid.AttrMap(self.btn_skip, 'dialog', 'dialog_focus')
		self.btn_rename_existing = urwid.RadioButton(bgroup, 'Rename Existing')
		attr_btn_rename_existing = urwid.AttrMap(self.btn_rename_existing, 'dialog', 'dialog_focus')
		self.btn_rename_copy = urwid.RadioButton(bgroup, 'Rename Copy')
		attr_btn_rename_copy = urwid.AttrMap(self.btn_rename_copy, 'dialog', 'dialog_focus')
		self.btn_rename_existing.set_state(True)
		w = urwid.SimpleFocusListWalker([
			label,
			attr_btn_overwrite,
			attr_btn_skip,
			attr_btn_rename_existing,
			attr_btn_rename_copy,
		])
		w = urwid.BoxAdapter(urwid.ListBox(w), 5)
		w = urwid.Columns([urwid.Filler(urwid.Divider(' ')), (19, urwid.Filler(w)), urwid.Filler(urwid.Divider(' '))])
		self.divider = urwid.LineBox(urwid.Padding(w, left=1, right=1), tlcorner='├', trcorner='┤', bline='')
		middle = urwid.Padding(self.divider, left=1, right=1)

		self.btn_ok = urwid.Button('OK', self.on_click_ok)
		attr_btn_ok = urwid.AttrMap(self.btn_ok, 'dialog', 'dialog_focus')
		self.btn_cancel = urwid.Button('Cancel', on_cancel)
		attr_btn_cancel = urwid.AttrMap(self.btn_cancel, 'dialog', 'dialog_focus')
		w = urwid.Columns([urwid.Divider(' '), (6, attr_btn_ok), (1, urwid.Text(' ')), (10, attr_btn_cancel), urwid.Divider(' ')])
		w = urwid.LineBox(urwid.Filler(w), tlcorner='├', trcorner='┤')
		bottom = urwid.Padding(w, left=1, right=1)

		self.pile = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			(3, top),
			(6, middle),
			(3, bottom),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w = urwid.AttrMap(self.pile, 'dialog')

		super().__init__(w)

	def keypress(self, size, key):
		if key in ('esc', 'f10'):
			self.btn_cancel.keypress(size, 'enter')
			return

		if self.pile.focus_position == 1:
			if key == 'tab':
				self.pile.focus_position = 2
			elif key == 'shift tab':
				self.pile.focus_position = 3
			elif key == 'up':
				pass
			elif key.startswith('ctrl ') or key.startswith('meta ') or key.startswith('shift '):
				pass
			elif key.startswith('f') and (len(key) > 1):
				pass
			elif key == 'backspace':
				if self.edit.edit_pos:
					return super().keypress(size, key)
			elif key == 'enter':
				self.btn_ok.keypress(size, 'enter')
				return
			else:
				return super().keypress(size, key)
		elif self.pile.focus_position == 2:
			if key == 'tab':
				self.pile.focus_position = 3
			elif key == 'shift tab':
				self.pile.focus_position = 1
			elif key in ('down', 'up', ' ', 'enter'):
				return super().keypress(size, key)
			elif key == 'j':
				return super().keypress(size, 'down')
			elif key == 'k':
				return super().keypress(size, 'up')
		elif self.pile.focus_position == 3:
			if key == 'tab':
				self.pile.focus_position = 1
			elif key == 'shift tab':
				self.pile.focus_position = 2
			elif key in ('left', 'up', 'right', ' ', 'enter'):
				return super().keypress(size, key)
			elif key == 'h':
				return super().keypress(size, 'left')
			elif key == 'k':
				return super().keypress(size, 'up')
			elif key == 'l':
				return super().keypress(size, 'right')

	def on_click_ok(self, button):
		if self.btn_overwrite.state:
			on_conflict = 'overwrite'
		elif self.btn_rename_existing.state:
			on_conflict = 'rename_existing'
		elif self.btn_rename_copy.state:
			on_conflict = 'rename_copy'
		else:
			on_conflict = 'skip'

		self.on_ok(apply_template(self.edit.get_edit_text(), self.controller.screen, quote=False), on_conflict)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import urwid

from .utils import (human_readable_size, TildeLayout)


class DlgDirscan(urwid.WidgetWrap):
	def __init__(self, controller, cwd, q, ev_abort, ev_skip, on_complete):
		self.controller = controller
		self.q = q
		self.ev_abort = ev_abort
		self.ev_skip = ev_skip
		self.on_complete = on_complete

		self.current = urwid.Text(cwd, layout=TildeLayout)
		self.files = urwid.Text(f'Files: 0', layout=TildeLayout)
		self.bytes = urwid.Text(f'Total size: 0', layout=TildeLayout)
		w = urwid.Pile([
			(1, urwid.Filler(self.current)),
			(1, urwid.Filler(self.files)),
			(1, urwid.Filler(self.bytes)),
		])
		w = urwid.LineBox(urwid.Padding(w, left=1, right=1), 'Directory scanning', title_attr='dialog_title', bline='')
		top = urwid.Padding(w, left=1, right=1)

		self.btn_abort = urwid.Button('Abort', lambda x: self.on_abort())
		attr_btn_abort = urwid.AttrMap(self.btn_abort, 'dialog', 'dialog_focus')
		self.btn_skip = urwid.Button('Skip', lambda x: self.on_skip())
		attr_btn_skip = urwid.AttrMap(self.btn_skip, 'dialog', 'dialog_focus')
		w = urwid.Columns([urwid.Divider(' '), (9, attr_btn_abort), (1, urwid.Text(' ')), (8, attr_btn_skip), urwid.Divider(' ')])
		w = urwid.LineBox(urwid.Filler(w), tlcorner='├', trcorner='┤')
		bottom = urwid.Padding(w, left=1, right=1)

		w = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			(4, top),
			(3, bottom),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w = urwid.AttrMap(w, 'dialog')

		super().__init__(w)

	def keypress(self, size, key):
		if key in ('left', 'right', ' ', 'enter'):
			return super().keypress(size, key)
		elif key == 'h':
			return super().keypress(size, 'left')
		elif key == 'l':
			return super().keypress(size, 'right')

	def on_pipe_data(self, data):
		retval = None
		info = None
		while not self.q.empty():
			info = self.q.get()

		if not info:
			pass
		elif 'result' in info:
			retval = False
			self.controller.screen.close_dialog()
			if not self.ev_abort.is_set():
				self.on_complete(info['result'], info['error'], info['skipped'])
		else:
			self.current.set_text(info['current'])
			self.files.set_text(f'Files: {info["files"]}')
			self.bytes.set_text(f'Total size: {human_readable_size(info["bytes"])}')

		return retval

	def on_abort(self):
		self.ev_abort.set()
		self.controller.screen.close_dialog()

		return False

	def on_skip(self):
		self.ev_skip.set()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import string
import collections

from atomicwrites import atomic_write


BOOKMARK_KEYS = tuple(string.ascii_letters + string.digits)


class Bookmarks(collections.UserDict):
	def __init__(self, file):
		super().__init__()

		self.file = str(file)

		try:
			with open(self.file, 'r') as fh:
				for line in fh:
					line = line.rstrip('\r\n')
					if (len(line) < 3) or (line[1] != ':') or (line[0] not in BOOKMARK_KEYS):
						continue

					self.data[line[0]] = line[2:]
		except (FileNotFoundError, PermissionError, IsADirectoryError):
			pass

	def __setitem__(self, key, item):
		key = str(key)
		item = str(item)

		if (len(key) != 1) or (key not in BOOKMARK_KEYS):
			return

		if (key not in self.data) or (self.data[key] != item):
			self.data[key] = item

			try:
				with atomic_write(self.file, overwrite=True) as f:
					for k, v in sorted(self.data.items()):
						f.write(f'{k}:{v}\n')
			except (FileNotFoundError, PermissionError, IsADirectoryError):
				pass




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

from pprint import pprint


debug_fh = None


def debug_print(*args, **kwargs):
	if debug_fh:
		print(*args, **kwargs, file=debug_fh)

def debug_pprint(*args, **kwargs):
	if debug_fh:
		pprint(*args, **kwargs, stream=debug_fh)

def set_debug_fh(fh):
	global debug_fh
	debug_fh = fh




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

from pathlib import Path

import xdg.BaseDirectory

CONFIG_DIR = Path(xdg.BaseDirectory.save_config_path('rnr'))

sys.path.insert(0, str(CONFIG_DIR))
_dont_write_bytecode = sys.dont_write_bytecode
sys.dont_write_bytecode = True

from .config import *

try:
	from config import *
except ModuleNotFoundError:
	try:
		shutil.copy(Path(__file__).parent / 'config.py', CONFIG_DIR)
		from config import *
	except (ModuleNotFoundError, FileNotFoundError, PermissionError, IsADirectoryError):
		pass

sys.dont_write_bytecode = _dont_write_bytecode
del _dont_write_bytecode
sys.path.pop(0)




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import subprocess
import signal

from pathlib import Path

import urwid

from .utils import (tar_stem, tar_suffix, apply_template)
from .debug_print import (debug_print, debug_pprint)



class CmdEdit(urwid.Edit):
	def keypress(self, size, key):
		if key in ('up', 'down'):
			pass
		elif key == 'backspace':
			if self.edit_pos:
				return super().keypress(size, key)
		else:
			return super().keypress(size, key)


class CmdBar(urwid.WidgetWrap):
	def __init__(self, controller, screen):
		self.controller = controller
		self.screen = screen

		self.action = None
		self.leader = ''
		self.file = None
		self.callback = None
		self.forced_focus = False

		self.edit = CmdEdit()

		w = urwid.AttrMap(self.edit, 'default')

		super().__init__(w)

		urwid.connect_signal(self.edit, 'change', self.on_change)

	def on_change(self, edit, new_edit_text):
		if self.action == 'filter':
			self.screen.center.focus.filter(new_edit_text)
			self.screen.center.focus.force_focus()

	def reset(self):
		self.edit.set_caption('')
		self.edit.set_edit_text('')
		self.screen.pile.focus_position = 0
		if self.forced_focus:
			self.screen.center.focus.remove_force_focus()

		self.action = None
		self.leader = ''
		self.file = None
		self.callback = None
		self.forced_focus = False

	def execute(self):
		if self.action == 'mkdir':
			self.do_mkdir()
		elif self.action == 'rename':
			self.do_rename()
		elif self.action == 'tag_glob':
			self.do_tag_glob()
		elif self.action == 'untag_glob':
			self.do_untag_glob()
		elif self.action == 'shell':
			self.do_shell()
		elif self.action == 'save':
			self.do_save()

		self.action = None
		self.leader = ''

		self.edit.set_caption('')
		self.edit.set_edit_text('')
		self.screen.pile.focus_position = 0
		if self.forced_focus:
			self.screen.center.focus.remove_force_focus()

		self.forced_focus = False

		if self.callback:
			file = self.file
			callback = self.callback
			self.file = None
			self.callback = None
			callback(file)


	def set_leader(self, leader):
		self.leader = leader
		self.edit.set_caption(self.leader)

	def prepare_action(self, action, caption, text, edit_pos=-1, forced_focus=True):
		self.action = action
		self.edit.set_caption(caption)
		self.edit.set_edit_text(text)
		if edit_pos < 0:
			self.edit.set_edit_pos(len(text))
		else:
			self.edit.set_edit_pos(edit_pos)
		self.screen.pile.focus_position = 1
		self.forced_focus = forced_focus
		if self.forced_focus:
			self.screen.center.focus.force_focus()

	def filter(self):
		self.prepare_action('filter', 'filter: ', self.screen.center.focus.file_filter)

	def mkdir(self, cwd):
		self.file = cwd
		self.prepare_action('mkdir', 'mkdir: ', '')

	def do_mkdir(self):
		new_dir = Path(apply_template(self.edit.get_edit_text(), self.screen, quote=False))
		try:
			new_dir = new_dir.expanduser()
		except RuntimeError:
			pass

		if new_dir.is_absolute():
			new_dir = Path(os.path.normpath(new_dir))
		else:
			new_dir = Path(os.path.normpath(self.file / new_dir))

		try:
			os.makedirs(new_dir, exist_ok=True)
			self.controller.reload(new_dir, only_focused=True)
		except (PermissionError, FileExistsError) as e:
			self.screen.error(f'{e.strerror} ({e.errno})')

	def rename(self, file, mode):
		self.file = file
		text = file.name.replace('%', '%%')
		if mode == 'replace':
			text = ''
			edit_pos = -1
		elif mode == 'insert':
			edit_pos = 0
		elif mode == 'append_before':
			edit_pos = len(tar_stem(file).replace('%', '%%'))
		elif mode == 'replace_before':
			text = tar_suffix(file).replace('%', '%%')
			edit_pos = 0
		else:
			edit_pos = -1

		self.prepare_action('rename', 'rename: ', text, edit_pos)

	def do_rename(self):
		new_name = Path(apply_template(self.edit.get_edit_text(), self.screen, quote=False))
		try:
			new_name = new_name.expanduser()
		except RuntimeError:
			pass

		if new_name.is_absolute():
			new_name = Path(os.path.normpath(new_name))
		else:
			new_name = Path(os.path.normpath(self.file.parent / new_name))

		try:
			if os.path.lexists(new_name):
				if new_name.is_dir():
					if new_name.resolve() == self.file.parent.resolve():
						return

					new_name = new_name / self.file.name
				else:
					if (new_name.parent.resolve() / new_name.name) == (self.file.parent.resolve() / self.file.name):
						return

					self.screen.error(f'File already exists')
					return

			self.file.rename(new_name)
			self.controller.reload(new_name, old_focus=self.file)
		except OSError as e:
			self.screen.error(f'{e.strerror} ({e.errno})')

	def tag_glob(self):
		self.prepare_action('tag_glob', 'tag: ', '*')

	def do_tag_glob(self):
		self.screen.center.focus.tag_glob(self.edit.get_edit_text())

	def untag_glob(self):
		self.prepare_action('untag_glob', 'untag: ', '*')

	def do_untag_glob(self):
		self.screen.center.focus.untag_glob(self.edit.get_edit_text())

	def shell(self):
		self.prepare_action('shell', 'shell: ', '')

	def do_shell(self):
		cwd = str(self.screen.center.focus.cwd)

		self.controller.loop.stop()
		prompt = ('$' if os.geteuid() else '#')
		cmd = apply_template(self.edit.get_edit_text(), self.screen)
		print(f'[{cwd}]{prompt} {cmd}')
		subprocess.run(cmd, shell=True, cwd=cwd)
		self.controller.loop.start()
		os.kill(os.getpid(), signal.SIGWINCH)
		self.controller.reload()

	def save(self, file, callback, forced_focus=True):
		self.file = file
		self.callback = callback
		text = str(file).replace('%', '%%')
		edit_pos = len(str(file.parent).replace('%', '%%'))
		self.prepare_action('save', 'save: ', text, edit_pos, forced_focus=forced_focus)

	def do_save(self):
		file = Path(apply_template(self.edit.get_edit_text(), self.screen, quote=False))
		try:
			file = file.expanduser()
		except RuntimeError:
			pass

		if file.is_absolute():
			file = Path(os.path.normpath(file))
		else:
			file = Path(os.path.normpath(self.file.parent / file))

		self.file = file

	def error(self, e):
		self.edit.set_caption(('default_error', f'ERROR: {e}'))
		self.edit.set_edit_text('')




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import re
import stat
import pwd
import grp
import functools
import collections
import shutil
import subprocess
import signal
import unicodedata

from pathlib import Path

import urwid

from fuzzyfinder import fuzzyfinder

from .utils import (human_readable_size, format_date, tar_stem, tar_suffix, TildeLayout, TLineWidget)
from .debug_print import (debug_print, debug_pprint)


ReNumbers = re.compile(r'(\d+)')

def try_int(s):
	try:
		return ('0', int(s))
	except ValueError:
		return (s, 0)

def natsort_key(s):
	return [try_int(x) for x in ReNumbers.split(unicodedata.normalize('NFKD', s.casefold()))]

def sort_by_name(a, b, reverse=False):
	if stat.S_ISDIR(a['stat'].st_mode) and (not stat.S_ISDIR(b['stat'].st_mode)):
		return (1 if reverse else -1)
	elif (not stat.S_ISDIR(a['stat'].st_mode)) and stat.S_ISDIR(b['stat'].st_mode):
		return (-1 if reverse else 1)
	elif a['key'] < b['key']:
		return -1
	elif a['key'] > b['key']:
		return 1
	else:
		return 0

def sort_by_extension(a, b, reverse=False):
	if stat.S_ISDIR(a['stat'].st_mode) and (not stat.S_ISDIR(b['stat'].st_mode)):
		return (1 if reverse else -1)
	elif (not stat.S_ISDIR(a['stat'].st_mode)) and stat.S_ISDIR(b['stat'].st_mode):
		return (-1 if reverse else 1)
	elif a['extension'] < b['extension']:
		return -1
	elif a['extension'] > b['extension']:
		return 1
	else:
		return sort_by_name(a, b, reverse)

def sort_by_date(a, b, reverse=False):
	if stat.S_ISDIR(a['stat'].st_mode) and (not stat.S_ISDIR(b['stat'].st_mode)):
		return (1 if reverse else -1)
	elif (not stat.S_ISDIR(a['stat'].st_mode)) and stat.S_ISDIR(b['stat'].st_mode):
		return (-1 if reverse else 1)
	elif a['lstat'].st_mtime < b['lstat'].st_mtime:
		return -1
	elif a['lstat'].st_mtime > b['lstat'].st_mtime:
		return 1
	else:
		return sort_by_name(a, b, reverse)

def sort_by_size(a, b, reverse=False):
	if stat.S_ISDIR(a['stat'].st_mode) and (not stat.S_ISDIR(b['stat'].st_mode)):
		return (1 if reverse else -1)
	elif (not stat.S_ISDIR(a['stat'].st_mode)) and stat.S_ISDIR(b['stat'].st_mode):
		return (-1 if reverse else 1)
	elif a['length'] < b['length']:
		return -1
	elif a['length'] > b['length']:
		return 1
	else:
		return sort_by_name(a, b, reverse)


def get_file_list(cwd):
	cwd = Path(cwd)

	uid_cache = Cache(lambda x: pwd.getpwuid(x).pw_name)
	gid_cache = Cache(lambda x: grp.getgrgid(x).gr_name)

	files = []
	for file in cwd.iterdir():
		obj = {
			'file': file,
			'key': natsort_key(file.name),
			'extension': natsort_key(tar_suffix(file)),
		}

		try:
			lstat = file.lstat()
		except FileNotFoundError:
			continue

		obj['lstat'] = lstat

		if stat.S_ISLNK(lstat.st_mode):
			try:
				st = file.stat()
				if stat.S_ISDIR(st.st_mode):
					obj['label'] = f'~{file.name}'
					obj['palette'] = 'dir_symlink'
				else:
					obj['label'] = f'@{file.name}'
					obj['palette'] = 'symlink'
			except (FileNotFoundError, PermissionError):
				st = lstat
				obj['label'] = f'!{file.name}'
				obj['palette'] = 'stalelink'
		else:
			st = lstat
			if stat.S_ISDIR(st.st_mode):
				obj['label'] = f'/{file.name}'
				obj['palette'] = 'directory'
			elif stat.S_ISCHR(lstat.st_mode):
				obj['label'] = f'-{file.name}'
				obj['palette'] = 'device'
			elif stat.S_ISBLK(lstat.st_mode):
				obj['label'] = f'+{file.name}'
				obj['palette'] = 'device'
			elif stat.S_ISFIFO(lstat.st_mode):
				obj['label'] = f'|{file.name}'
				obj['palette'] = 'special'
			elif stat.S_ISSOCK(lstat.st_mode):
				obj['label'] = f'={file.name}'
				obj['palette'] = 'special'
			elif lstat.st_mode & (stat.S_IXUSR | stat.S_IXGRP | stat.S_IXOTH):
				obj['label'] = f'*{file.name}'
				obj['palette'] = 'executable'
			else:
				obj['label'] = f' {file.name}'
				obj['palette'] = 'panel'

		obj['stat'] = st

		if stat.S_ISDIR(st.st_mode):
			try:
				length =len(list(file.iterdir()))
				obj['length'] = (length,)
				obj['size'] = str(length)
			except (FileNotFoundError, PermissionError):
				obj['length'] = (-1,)
				obj['size'] = '?'
		elif stat.S_ISCHR(lstat.st_mode) or stat.S_ISBLK(lstat.st_mode):
				major = os.major(lstat.st_rdev)
				minor = os.minor(lstat.st_rdev)
				obj['length'] = (major, minor)
				obj['size'] = f'{major},{minor}'
		else:
			length = lstat.st_size
			obj['length'] = (length,)
			obj['size'] = human_readable_size(length)

		try:
			uid = uid_cache[lstat.st_uid]
		except KeyError:
			uid = str(lstat.st_uid)

		try:
			gid = gid_cache[lstat.st_gid]
		except KeyError:
			gid = str(lstat.st_gid)

		obj['details'] = f'{stat.filemode(lstat.st_mode)} {lstat.st_nlink} {uid} {gid}'

		if stat.S_ISLNK(lstat.st_mode):
			try:
				link_target = os.readlink(file)

				obj['details'] = f'{obj["details"]} -> {link_target}'
				if Path(link_target).is_absolute():
					obj['link_target'] = Path(os.path.normpath(link_target))
				else:
					obj['link_target'] = Path(os.path.normpath(file.parent / link_target))
			except (FileNotFoundError, PermissionError):
				obj['details'] = f'{obj["details"]} -> ?'
				obj['link_target'] = file
		else:
			obj['details'] = f'{obj["details"]} {file.name}'

		files.append(obj)

	return files


class Cache(collections.defaultdict):
	def __missing__(self, key):
		if self.default_factory is None:
			raise KeyError(key)

		self[key] = value = self.default_factory(key)

		return value


class SelectableColumns(urwid.Columns):
	def __init__(self, *args, **kwargs):
		super().__init__(*args, **kwargs)

		self._selectable = True

	def keypress(self, size, key):
		return key


class VimListBox(urwid.ListBox):
	def keypress(self, size, key):
		if self.controller.leader:
			return key

		if key in ('h', 'left'):
			self.model.chdir(self.model.cwd.parent)
		elif key in ('j', 'down'):
			retval = None

			try:
				if (self.focus_position + 1) < len(self.model.walker):
					retval = super().keypress(size, 'down')
			except IndexError:
				pass

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('k', 'up'):
			retval = None

			try:
				if self.focus_position > 0:
					retval = super().keypress(size, 'up')
			except IndexError:
				pass

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('l', 'right', 'enter'):
			try:
				self.model.execute(self.model.walker.get_focus()[0].model)
			except AttributeError:
				pass
		elif key in ('g', 'home'):
			retval = super().keypress(size, 'home')

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('G', 'end'):
			retval = super().keypress(size, 'end')

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('ctrl b', 'page up'):
			retval = super().keypress(size, 'page up')

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('ctrl f', 'page down'):
			retval = super().keypress(size, 'page down')

			try:
				self.model.show_details(self.model.walker.get_focus()[0].model)
				self.model.show_preview(self.model.walker.get_focus()[0].model)
			except AttributeError:
				self.model.show_details(None)
				self.model.show_preview(None)

			return retval
		elif key in ('v', 'f3'):
			try:
				self.model.view(self.model.walker.get_focus()[0].model)
			except AttributeError:
				pass
		elif key in ('e', 'f4'):
			try:
				self.model.edit(self.model.walker.get_focus()[0].model)
			except AttributeError:
				pass
		elif key in ('insert', ' '):
			if self.model.tag_toggle():
				if (self.focus_position + 1) < len(self.model.walker):
					super().keypress(size, 'down')
		elif key == '*':
			self.model.tag_toggle_all()
		else:
			return super().keypress(size, key)


class Panel(urwid.WidgetWrap):
	def __init__(self, controller):
		self.controller = controller

		self.title = TLineWidget(urwid.Text('', layout=TildeLayout), title_align='left', lcorner='┌', rcorner='┐')
		title = urwid.AttrMap(self.title, 'panel')

		self.walker = urwid.SimpleFocusListWalker([])
		self.listbox = VimListBox(self.walker)
		self.listbox.model = self
		self.listbox.controller = controller
		listbox = urwid.LineBox(self.listbox, tline='', bline='')
		listbox = urwid.AttrMap(listbox, 'panel')

		txt_loading = urwid.LineBox(urwid.Filler(urwid.Text('Loading...'), valign='top'), tline='', bline='')
		self.txt_loading = urwid.AttrMap(txt_loading, 'panel')

		self.details_separator = TLineWidget(urwid.Text('─', layout=TildeLayout))
		details_separator = urwid.AttrMap(self.details_separator, 'panel')

		self.details = urwid.Text(' ', layout=TildeLayout)
		details = urwid.LineBox(self.details, tline='', bline='')
		details = urwid.AttrMap(details, 'panel')

		self.footer = TLineWidget(urwid.Text('', layout=TildeLayout), title_align='right', lcorner='└', rcorner='┘')
		footer = urwid.AttrMap(self.footer, 'panel')

		self.pile = urwid.Pile([('pack', title), listbox, ('pack', details_separator), ('pack', details), ('pack', footer)])

		cwd = Path.cwd()
		self.old_cwd = cwd
		self.cwd = cwd
		self.show_hidden = False
		self.sort_method = 'sort_by_name'
		self.reverse = False
		self.file_filter = ''
		self.forced_focus = None
		self.files = []
		self.shown_files = []
		self.filtered_files = []
		self.tagged_files = set()

		self.chdir(cwd)
		self.walker.set_focus(0)

		self.focused = False

		super().__init__(self.pile)

	def chdir(self, cwd, focus_path=None):
		self.title.set_title(f' {str(cwd)} ')
		self.file_filter = ''
		self.tagged_files.clear()
		self.update_tagged_count()

		if focus_path is None:
			focus_path = self.cwd

		if self._reload(cwd, focus_path):
			self.old_cwd = self.cwd
			self.cwd = cwd
			return True
		else:
			self.title.set_title(f' {str(self.cwd)} ')
			return False

	def reload(self, focus_path=None):
		files_to_discard = []
		for file in self.tagged_files:
			if not os.path.lexists(file):
				files_to_discard.append(file)

		for file in files_to_discard:
			self.tagged_files.discard(file)

		self.update_tagged_count()

		try:
			obj, focus_position = self.walker.get_focus()
			if focus_path is None:
				focus_path = obj.model['file']
		except AttributeError:
			focus_path = None
			focus_position = 0

		if not self._reload(self.cwd, focus_path, focus_position):
			old_cwd = self.old_cwd
			parent = self.cwd
			while parent.parent != parent:
				parent = parent.parent
				if self.chdir(parent):
					self.old_cwd = old_cwd
					break

	def _reload(self, cwd, focus_path, focus_position=0):
		cwd = Path(cwd)

		listbox = self.pile.contents[1]
		self.pile.contents[1] = (self.txt_loading, self.pile.options())
		if self.controller.loop:
			self.controller.loop.draw_screen()

		self.pile.contents[1] = listbox

		try:
			files = get_file_list(cwd)

		except (FileNotFoundError, PermissionError):
			return False

		self.files = files
		self.apply_hidden(self.show_hidden)
		self.apply_filter(self.file_filter)
		self.update_list_box(focus_path, focus_position)
		self.footer.set_title(f' Free: {human_readable_size(shutil.disk_usage(cwd).free)} ')

		return True

	def update_list_box(self, focus_path, focus_position=0):
		self.filtered_files.sort(key=functools.cmp_to_key(functools.partial(globals()[self.sort_method], reverse=self.reverse)), reverse=self.reverse)

		focus = -1
		labels = []
		for file in self.filtered_files:
			if file['file'] in self.tagged_files:
				attr_map = 'marked'
				focus_map = 'markselect'
			else:
				attr_map = file['palette']
				focus_map = 'selected'

			w = urwid.AttrMap(SelectableColumns([urwid.Text(file['label'], layout=TildeLayout), ('pack', urwid.Text(file['size'])), ('pack', urwid.Text(format_date(file['lstat'].st_mtime)))], dividechars=1), attr_map, focus_map)
			w.model = file

			if file['file'] == focus_path:
				focus = len(labels)

			labels.append(w)

		self.walker[:] = labels

		if focus < 0:
			focus = min(focus_position, len(labels) - 1)

		self.walker.set_focus(focus)

		try:
			self.show_details(self.filtered_files[focus])
		except IndexError:
			self.show_details(None)

		try:
			self.show_preview(self.filtered_files[focus])
		except IndexError:
			self.show_preview(None)
		except AttributeError:
			pass

	def apply_hidden(self, show_hidden):
		self.show_hidden = show_hidden

		if show_hidden:
			self.shown_files = self.files[:]
		else:
			self.shown_files = [x for x in self.files if not x['file'].name.startswith('.')]

	def apply_filter(self, filter):
		self.file_filter = filter

		if filter:
			self.filtered_files = list(fuzzyfinder(filter, self.shown_files, accessor=lambda x: x['file'].name))
		else:
			self.filtered_files = self.shown_files[:]

	def execute(self, file):
		if stat.S_ISDIR(file['stat'].st_mode):
			self.chdir(file['file'])
		elif 'link_target' in file:
			try:
				if file['link_target'].is_dir():
					if file['link_target'] != self.cwd:
						self.chdir(file['link_target'])
				elif os.path.lexists(file['link_target']):
					if file['link_target'].parent == self.cwd:
						for (i, line) in enumerate(self.walker):
							if line.model['file'] == file['link_target']:
								self.walker.set_focus(i)
								break
					else:
						self.chdir(file['link_target'].parent, file['link_target'])
			except (FileNotFoundError, PermissionError):
				pass
		else:
			self.controller.loop.stop()
			subprocess.run([self.controller.opener, file['file'].name], cwd=self.cwd)
			self.controller.loop.start()
			os.kill(os.getpid(), signal.SIGWINCH)
			self.controller.reload()

	def view(self, file):
		if stat.S_ISDIR(file['stat'].st_mode):
			self.chdir(file['file'])
		elif self.controller.use_internal_viewer:
			self.controller.view(file['file'])
		else:
			self.controller.loop.stop()
			subprocess.run([self.controller.pager, file['file'].name], cwd=self.cwd)
			self.controller.loop.start()
			os.kill(os.getpid(), signal.SIGWINCH)
			self.controller.reload()

	def edit(self, file):
		self.controller.loop.stop()
		subprocess.run([self.controller.editor, file['file'].name], cwd=self.cwd)
		self.controller.loop.start()
		os.kill(os.getpid(), signal.SIGWINCH)
		self.controller.reload()

	def set_title_attr(self, attr):
		self.title.set_title_attr(attr)

	def show_details(self, file):
		if file:
			self.details.set_text(file['details'])
		else:
			self.details.set_text(' ')

	def show_preview(self, file):
		if not (self.controller.screen.show_preview and self.focused):
			return

		if file:
			self.controller.screen.preview_panel.set_title(file['file'].name)

			if stat.S_ISDIR(file['stat'].st_mode):
				self.controller.screen.preview_panel.read_directory(file['file'])
			elif stat.S_ISREG(file['stat'].st_mode):
				self.controller.screen.preview_panel.read_file(file['file'], file['stat'].st_size)
			else:
				self.controller.screen.preview_panel.clear()
		else:
			self.controller.screen.preview_panel.clear_title()
			self.controller.screen.preview_panel.clear()

	def filter(self, filter):
		if not filter:
			try:
				focus_path = self.get_focus()['file']
			except TypeError:
				focus_path = None

		self.apply_filter(filter)

		if filter:
			try:
				focus_path = self.filtered_files[0]['file']
			except IndexError:
				focus_path = None

		self.update_list_box(focus_path)

	def get_focus(self):
		try:
			return self.walker.get_focus()[0].model
		except AttributeError:
			return None

	def force_focus(self):
		self.remove_force_focus()

		self.forced_focus = self.walker.get_focus()[0]

		try:
			file = self.forced_focus.model

			if file['file'] in self.tagged_files:
				attr_map = 'markselect'
			else:
				attr_map = 'selected'

			self.forced_focus.set_attr_map({None: attr_map})
		except AttributeError:
			pass

	def remove_force_focus(self):
		try:
			file = self.forced_focus.model

			if file['file'] in self.tagged_files:
				attr_map = 'marked'
			else:
				attr_map = file['palette']

			self.forced_focus.set_attr_map({None: attr_map})
		except AttributeError:
			pass

		self.forced_focus = None

	def toggle_hidden(self):
		self.apply_hidden(not self.show_hidden)
		self.apply_filter(self.file_filter)

		try:
			focus_path = self.get_focus()['file']
		except TypeError:
			focus_path = None

		self.update_list_box(focus_path)

	def sort(self, sort_method, reverse=False):
		self.sort_method = sort_method
		self.reverse = reverse

		try:
			focus_path = self.get_focus()['file']
		except TypeError:
			focus_path = None

		self.update_list_box(focus_path)

	def update_tagged_count(self):
		if self.tagged_files:
			self.details_separator.set_title(f' {len(self.tagged_files)} File(s) Tagged ')
			self.details_separator.set_title_attr('marked')
		else:
			self.details_separator.set_title('─')
			self.details_separator.set_title_attr('panel')

	def tag_toggle(self):
		try:
			line = self.walker.get_focus()[0]
			file = line.model['file']
		except AttributeError:
			return False

		if file in self.tagged_files:
			self.tagged_files.discard(file)
			line.set_attr_map({None: line.model['palette']})
			line.set_focus_map({None: 'selected'})
		else:
			self.tagged_files.add(file)
			line.set_attr_map({None: 'marked'})
			line.set_focus_map({None: 'markselect'})

		self.update_tagged_count()

		return True

	def tag_toggle_all(self):
		for line in self.walker:
			file = line.model['file']

			if file in self.tagged_files:
				self.tagged_files.discard(file)
				line.set_attr_map({None: line.model['palette']})
				line.set_focus_map({None: 'selected'})
			else:
				self.tagged_files.add(file)
				line.set_attr_map({None: 'marked'})
				line.set_focus_map({None: 'markselect'})

		self.update_tagged_count()

	def tag_glob(self, pattern):
		try:
			for line in self.walker:
				file = line.model['file']

				if file.match(pattern):
					self.tagged_files.add(file)
					line.set_attr_map({None: 'marked'})
					line.set_focus_map({None: 'markselect'})
		except ValueError:
			pass

		self.update_tagged_count()

	def untag_glob(self, pattern):
		try:
			for line in self.walker:
				file = line.model['file']

				if file.match(pattern):
					self.tagged_files.discard(file)
					line.set_attr_map({None: line.model['palette']})
					line.set_focus_map({None: 'selected'})
		except ValueError:
			pass

		self.update_tagged_count()

	def untag_all(self):
		for line in self.walker:
			file = line.model['file']

			self.tagged_files.discard(file)
			line.set_attr_map({None: line.model['palette']})
			line.set_focus_map({None: 'selected'})

		self.update_tagged_count()

	def get_tagged_files(self):
		if self.tagged_files:
			return sorted(self.tagged_files, key=lambda x: natsort_key(x.name))
		else:
			try:
				file = self.get_focus()['file']
			except TypeError:
				file = None

			if file:
				return [file]
			else:
				return []




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import urwid

from .utils import (human_readable_size, format_seconds, TildeLayout)


class DlgDeleteProgress(urwid.WidgetWrap):
	def __init__(self, controller, num_files, total_size, q, ev_skip, ev_suspend, ev_abort, ev_nodb, on_complete):
		self.controller = controller
		self.num_files = num_files
		self.total_size = total_size
		self.q = q
		self.ev_skip = ev_skip
		self.ev_suspend = ev_suspend
		self.ev_abort = ev_abort
		self.ev_nodb = ev_nodb
		self.on_complete = on_complete

		self.aborted = False

		self.current = urwid.Text(' ', layout=TildeLayout)
		w = urwid.Filler(self.current)
		w = urwid.LineBox(urwid.Padding(w, left=1, right=1), 'Delete', title_attr='dialog_title', bline='')
		top = urwid.Padding(w, left=1, right=1)

		self.files = urwid.Text(f'Files processed: 0/{self.num_files}', layout=TildeLayout)
		self.time = urwid.Text(f'Time: {format_seconds(0)} ETA {format_seconds(0)}', layout=TildeLayout)
		self.progress = urwid.ProgressBar('dialog', 'progress', 0, (num_files or 1))
		w = urwid.Columns([(1, urwid.Text('[')), self.progress, (1, urwid.Text(']'))])
		w = urwid.Pile([
			(1, urwid.Filler(w)),
			(1, urwid.Filler(self.files)),
			(1, urwid.Filler(self.time)),
		])
		self.divider = urwid.LineBox(urwid.Padding(w, left=1, right=1), f'Total: {human_readable_size(0)}/{human_readable_size(self.total_size)}', tlcorner='├', trcorner='┤', bline='')
		middle = urwid.Padding(self.divider, left=1, right=1)

		self.btn_skip = urwid.Button('Skip', lambda x: self.on_skip())
		attr_btn_skip = urwid.AttrMap(self.btn_skip, 'dialog', 'dialog_focus')
		self.btn_suspend = urwid.Button('Suspend', lambda x: self.on_suspend())
		attr_btn_suspend = urwid.AttrMap(self.btn_suspend, 'dialog', 'dialog_focus')
		self.btn_abort = urwid.Button('Abort', lambda x: self.on_abort())
		attr_btn_abort = urwid.AttrMap(self.btn_abort, 'dialog', 'dialog_focus')
		self.btn_nodb = urwid.Button('No DB', lambda x: self.on_nodb())
		attr_btn_nodb = urwid.AttrMap(self.btn_nodb, 'dialog', 'dialog_focus')
		w = urwid.Columns([urwid.Divider(' '), (8, attr_btn_skip), (1, urwid.Text(' ')), (12, attr_btn_suspend), (1, urwid.Text(' ')), (9, attr_btn_abort), (1, urwid.Text(' ')), (9, attr_btn_nodb), urwid.Divider(' ')])
		w = urwid.LineBox(urwid.Filler(w), tlcorner='├', trcorner='┤')
		bottom = urwid.Padding(w, left=1, right=1)

		w = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			(2, top),
			(4, middle),
			(3, bottom),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w = urwid.AttrMap(w, 'dialog')

		super().__init__(w)

	def keypress(self, size, key):
		if key in ('left', 'right', ' ', 'enter'):
			return super().keypress(size, key)
		elif key == 'h':
			return super().keypress(size, 'left')
		elif key == 'l':
			return super().keypress(size, 'right')

	def on_pipe_data(self, data):
		retval = None
		info = None
		while not self.q.empty():
			info = self.q.get()

		if not info:
			pass
		elif 'result' in info:
			retval = False
			self.controller.screen.close_dialog()
			self.controller.suspend.discard(self.ev_suspend)
			self.on_complete(info['result'], info['error'], info['skipped'], info['aborted'])
		else:
			self.current.set_text(info['current'])
			self.divider.set_title(f'Total: {human_readable_size(info["bytes"])}/{human_readable_size(self.total_size)}')
			self.files.set_text(f'Files processed: {info["files"]}/{self.num_files}')

			fps = info['files'] / (info['time'] or 1)
			eta = max(int(round((self.num_files - info['files']) / (fps or 1))), 0)
			self.time.set_text(f'Time: {format_seconds(info["time"])} ETA {format_seconds(eta)}')

			self.progress.set_completion(info['files'])

		return retval

	def on_skip(self):
		if self.aborted:
			return

		self.ev_skip.set()

		if not self.ev_suspend.is_set():
			self.on_suspend()

	def on_suspend(self):
		if self.aborted:
			return

		if self.ev_suspend.is_set():
			self.ev_suspend.clear()
			self.btn_suspend.set_label('Continue')
		else:
			self.ev_suspend.set()
			self.btn_suspend.set_label('Suspend')

	def on_abort(self):
		if self.aborted:
			return

		self.ev_abort.set()

		if not self.ev_suspend.is_set():
			self.on_suspend()

		self.aborted = True

	def on_nodb(self):
		self.ev_nodb.set()




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

from pathlib import Path

import urwid

from atomicwrites import atomic_write

from .database import DataBase
from .utils import (human_readable_size, format_seconds, TildeLayout)
from .debug_print import (debug_print, debug_pprint)


class DlgReport(urwid.WidgetWrap):
	def __init__(self, controller, completed_list, error_list, skipped_list, aborted_list, operation, files, cwd, dest, scan_error, scan_skipped, job_id):
		self.controller = controller
		self.command_bar = controller.screen.command_bar
		self.completed_list = completed_list
		self.error_list = error_list
		self.skipped_list = skipped_list
		self.aborted_list = aborted_list
		self.operation = operation
		self.files = files
		self.cwd = cwd
		self.dest = dest
		self.scan_error = scan_error
		self.scan_skipped = scan_skipped
		self.job_id = job_id

		if scan_error or error_list or aborted_list:
			attr = 'error'
			title_attr = 'error_title'
			focus_attr = 'error_focus'
		else:
			attr = 'dialog'
			title_attr = 'dialog_title'
			focus_attr = 'dialog_focus'

		self.messages = []
		self.messages.extend([f'ERROR (scan) [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in scan_error])
		self.messages.extend([f'SKIPPED (scan) [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in scan_skipped])
		self.messages.extend([f'ERROR [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in error_list])
		self.messages.extend([f'SKIPPED [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in skipped_list])
		if aborted_list:
			self.messages.extend([f'{("WARNING" if x["message"] else "DONE")} [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in completed_list])
			self.messages.extend([f'ABORTED [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in aborted_list])
		else:
			self.messages.extend([f'WARNING [{x["message"]}]: {str(Path(x["file"]).relative_to(cwd))}' for x in completed_list if x['message']])

		l = [urwid.Text(x, layout=TildeLayout) for x in self.messages]
		w = urwid.SimpleListWalker(l)
		self.listbox = urwid.ListBox(w)
		w = urwid.LineBox(urwid.Padding(self.listbox, left=1, right=1), 'Report', title_attr=title_attr, bline='')
		top = urwid.Padding(w, left=1, right=1)

		self.btn_close = urwid.Button('Close', lambda x: self.on_close())
		attr_btn_close = urwid.AttrMap(self.btn_close, attr, focus_attr)
		self.btn_save = urwid.Button('Save', lambda x: self.on_save())
		attr_btn_save = urwid.AttrMap(self.btn_save, attr, focus_attr)
		w = urwid.Columns([urwid.Divider(' '), (9, attr_btn_close), (1, urwid.Text(' ')), (8, attr_btn_save), urwid.Divider(' ')])
		w = urwid.LineBox(urwid.Filler(w), tlcorner='├', trcorner='┤')
		bottom = urwid.Padding(w, left=1, right=1)

		w = urwid.Pile([
			(1, urwid.Filler(urwid.Text(' '))),
			top,
			(3, bottom),
			(1, urwid.Filler(urwid.Text(' '))),
		])
		w.set_focus(2)
		w = urwid.AttrMap(w, attr)

		super().__init__(w)

	def keypress(self, size, key):
		if key in ('left', 'right', ' ', 'enter'):
			return super().keypress(size, key)
		elif key == 'h':
			return super().keypress(size, 'left')
		elif key == 'l':
			return super().keypress(size, 'right')
		elif key in ('j', 'down', 'ctrl f', 'page down'):
			self.listbox.keypress(size, 'page down')
		elif key in ('k', 'up', 'ctrl b', 'page up'):
			self.listbox.keypress(size, 'page up')
		elif key in ('g', 'home'):
			self.listbox.keypress(size, 'home')
		elif key in ('G', 'end'):
			self.listbox.keypress(size, 'end')

	def on_close(self):
		if self.controller.dbfile:
			db = DataBase(self.controller.dbfile)
			db.delete_job(self.job_id)
			del db

		self.controller.screen.close_dialog()
		self.command_bar.reset()

		if self.controller.pending_jobs:
			self.controller.show_next_pending_job()
		else:
			self.controller.reload()

	def on_save(self):
		self.command_bar.save(Path(self.cwd) / 'rnr-report.txt', self.do_save, forced_focus=False)

	def do_save(self, file):
		try:
			with atomic_write(str(file)) as f:
				f.write(f'Operation: {self.operation}\n')
				f.write(f'From: {str(self.cwd)}\n')
				if self.dest:
					f.write(f'To: {str(self.dest)}\n')

				f.write('Files:\n')
				for file in self.files:
					f.write(f'{str(file.relative_to(self.cwd))}\n')

				f.write('\n')
				f.write('------------------------------------------------------------------------------\n')
				f.write('\n')

				for message in self.messages:
					f.write(f'{message}\n')

			self.on_close()
		except OSError as e:
			self.command_bar.error(f'{e.strerror} ({e.errno})')




#! /usr/bin/env python3
# -*- coding: utf-8 -*-

# Copyright (C) 2020  Franco Bugnano
#
# This program is free software: you can redistribute it and/or modify
# it under the terms of the GNU General Public License as published by
# the Free Software Foundation, either version 3 of the License, or
# (at your option) any later version.
#
# This program is distributed in the hope that it will be useful,
# but WITHOUT ANY WARRANTY; without even the implied warranty of
# MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
# GNU General Public License for more details.
#
# You should have received a copy of the GNU General Public License
# along with this program.  If not, see <https://www.gnu.org/licenses/>.

import sys
import os

import re
import stat
import pwd
import grp
import functools
import collections
import shutil
import subprocess
import signal
import unicodedata

from pathlib import Path

import urwid

from fuzzyfinder import fuzzyfinder

from . import rnrview

from .utils import (human_readable_size, format_date, tar_stem, tar_suffix, TildeLayout, TLineWidget)
from .debug_print import (debug_print, debug_pprint)


class PreviewPanel(urwid.WidgetWrap):
	def __init__(self, controller):
		self.controller = controller

		self.title = TLineWidget(urwid.Text(' (Preview) ', layout=TildeLayout), title_align='left', lcorner='┌', rcorner='┐')
		title = urwid.AttrMap(self.title, 'panel')

		self.walker = urwid.SimpleFocusListWalker([])
		self.listbox = rnrview.FileViewListBox(controller, controller.tabsize, use_line_highlight=True)
		self.attr_listbox = urwid.AttrMap(self.listbox, 'Text')
		listbox = urwid.LineBox(self.attr_listbox, tline='', bline='')
		listbox = urwid.AttrMap(listbox, 'panel')

		self.footer = TLineWidget(urwid.Text('', layout=TildeLayout), title_align='right', lcorner='└', rcorner='┘')
		footer = urwid.AttrMap(self.footer, 'panel')

		self.pile = urwid.Pile([('pack', title), listbox, ('pack', footer)])

		self.focused = False

		self.cwd = Path.cwd()

		super().__init__(self.pile)

	def read_file(self, filename, file_size):
		try:
			self.attr_listbox.set_attr_map({None: 'Text'})
			self.listbox.read_file(filename, file_size)
			self.cwd = Path(filename).parent
		except OSError:
			self.listbox.clear()

	def read_directory(self, filename):
		try:
			self.attr_listbox.set_attr_map({None: 'panel'})
			self.listbox.read_directory(filename)
			self.cwd = Path(filename)
		except OSError:
			self.listbox.clear()

	def clear(self):
		self.listbox.clear()

	def set_title_attr(self, attr):
		self.title.set_title_attr(attr)

	def set_title(self, title):
		self.title.set_title(f' (Preview) {str(title)} ')

	def clear_title(self):
		self.title.set_title(f' (Preview) ')

